import pytest
from mau.lexers.arguments_lexer import ArgumentsLexer
from mau.nodes.arguments import NamedArgumentNode, UnnamedArgumentNode
from mau.parsers.arguments import set_names_and_defaults
from mau.parsers.arguments_parser import ArgumentsParser
from mau.parsers.base_parser import ParserError

from tests.helpers import parser_runner_factory

runner = parser_runner_factory(ArgumentsLexer, ArgumentsParser)


def test_single_unnamed_argument_no_spaces():
    source = "value1"

    assert runner(source).nodes == [UnnamedArgumentNode("value1")]


def test_single_unnamed_argument_with_spaces():
    source = "value with spaces"

    assert runner(source).nodes == [UnnamedArgumentNode("value with spaces")]


def test_multiple_unnamed_arguments_no_spaces():
    source = "value1,value2"

    assert runner(source).nodes == [
        UnnamedArgumentNode("value1"),
        UnnamedArgumentNode("value2"),
    ]


def test_multiple_unnamed_arguments_with_spaces():
    source = "value1 with spaces,value2 with more spaces"

    assert runner(source).nodes == [
        UnnamedArgumentNode("value1 with spaces"),
        UnnamedArgumentNode("value2 with more spaces"),
    ]


def test_multiple_unnamed_arguments_space_after_comma_is_removed():
    source = "value1 with spaces, value2 with more spaces"

    assert runner(source).nodes == [
        UnnamedArgumentNode("value1 with spaces"),
        UnnamedArgumentNode("value2 with more spaces"),
    ]


def test_multiple_unnamed_arguments_multiple_spaces_after_comma_is_removed():
    source = "value1 with spaces,    value2 with more spaces"

    assert runner(source).nodes == [
        UnnamedArgumentNode("value1 with spaces"),
        UnnamedArgumentNode("value2 with more spaces"),
    ]


def test_single_unnamed_argument_with_quotes_no_spaces():
    source = '"value1"'

    assert runner(source).nodes == [UnnamedArgumentNode("value1")]


def test_single_unnamed_argument_with_quotes_with_spaces():
    source = '"value with spaces"'

    assert runner(source).nodes == [UnnamedArgumentNode("value with spaces")]


def test_single_unnamed_argument_comma_is_ignored_between_quotes():
    source = '"value1,value2"'

    assert runner(source).nodes == [UnnamedArgumentNode("value1,value2")]


def test_multiple_unnamed_arguments_with_quotes_no_spaces():
    source = '"value1","value2"'

    assert runner(source).nodes == [
        UnnamedArgumentNode("value1"),
        UnnamedArgumentNode("value2"),
    ]


def test_multiple_unnamed_arguments_with_quotes_with_spaces():
    source = '"value1 with spaces","value2 with more spaces"'

    assert runner(source).nodes == [
        UnnamedArgumentNode("value1 with spaces"),
        UnnamedArgumentNode("value2 with more spaces"),
    ]


def test_single_unnamed_argument_with_non_delimiting_quotes():
    source = r'value "with quotes"'

    assert runner(source).nodes == [UnnamedArgumentNode('value "with quotes"')]


def test_single_unnamed_argument_with_escaped_quotes():
    source = r'"value \"with escaped quotes\""'

    assert runner(source).nodes == [UnnamedArgumentNode('value "with escaped quotes"')]


def test_single_named_argument():
    source = "name=value1"

    assert runner(source).nodes == [
        NamedArgumentNode("name", "value1"),
    ]


def test_single_named_argument_with_spaces():
    source = "name=value with spaces"

    assert runner(source).nodes == [
        NamedArgumentNode("name", "value with spaces"),
    ]


def test_multiple_named_arguments():
    source = "name1=value1,name2=value2"

    assert runner(source).nodes == [
        NamedArgumentNode("name1", "value1"),
        NamedArgumentNode("name2", "value2"),
    ]


def test_multiple_named_arguments_with_spaces():
    source = "name1=value1 with spaces,name2=value2 with spaces"

    assert runner(source).nodes == [
        NamedArgumentNode("name1", "value1 with spaces"),
        NamedArgumentNode("name2", "value2 with spaces"),
    ]


def test_multiple_named_arguments_space_after_comma_is_removed():
    source = "name1=value1, name2=value2"

    assert runner(source).nodes == [
        NamedArgumentNode("name1", "value1"),
        NamedArgumentNode("name2", "value2"),
    ]


def test_multiple_named_arguments_multiple_spaces_after_comma_is_removed():
    source = "name1=value1,    name2=value2"

    assert runner(source).nodes == [
        NamedArgumentNode("name1", "value1"),
        NamedArgumentNode("name2", "value2"),
    ]


def test_single_named_argument_with_quotes_no_spaces():
    source = 'name="value1"'

    assert runner(source).nodes == [NamedArgumentNode("name", "value1")]


def test_single_named_argument_with_quotes_with_spaces():
    source = 'name="value with spaces"'

    assert runner(source).nodes == [NamedArgumentNode("name", "value with spaces")]


def test_single_named_argument_comma_is_ignored_between_quotes():
    source = 'name="value1,value2"'

    assert runner(source).nodes == [NamedArgumentNode("name", "value1,value2")]


def test_multiple_named_arguments_with_quotes_no_spaces():
    source = 'name1="value1",name2="value2"'

    assert runner(source).nodes == [
        NamedArgumentNode("name1", "value1"),
        NamedArgumentNode("name2", "value2"),
    ]


def test_multiple_named_arguments_with_quotes_with_spaces():
    source = 'name1="value1 with spaces",name2="value2 with more spaces"'

    assert runner(source).nodes == [
        NamedArgumentNode("name1", "value1 with spaces"),
        NamedArgumentNode("name2", "value2 with more spaces"),
    ]


def test_single_named_argument_with_non_delimiting_quotes():
    source = r'name=value "with quotes"'

    assert runner(source).nodes == [NamedArgumentNode("name", 'value "with quotes"')]


def test_single_named_argument_with_escaped_quotes():
    source = r'name="value \"with escaped quotes\""'

    assert runner(source).nodes == [
        NamedArgumentNode("name", 'value "with escaped quotes"')
    ]


def test_unnamed_and_named_arguments():
    source = "value1, name=value2"

    assert runner(source).nodes == [
        UnnamedArgumentNode("value1"),
        NamedArgumentNode("name", "value2"),
    ]


def test_named_arguments_followed_by_unnamed():
    source = "name=value2, value1"

    with pytest.raises(ParserError):
        runner(source)


def test_apply_prototype_unnamed_arguments():
    args, kwargs = set_names_and_defaults(
        args=["value1", "value2"],
        kwargs={},
        positional_names=["attr1", "attr2"],
        default_values={"attr3": "value3"},
    )

    assert args == []
    assert kwargs == {"attr1": "value1", "attr2": "value2", "attr3": "value3"}


def test_apply_prototype_unnamed_and_named_arguments():
    parser = ArgumentsParser(tokens=[])

    parser.nodes = [
        UnnamedArgumentNode("value1"),
        UnnamedArgumentNode("value2"),
        NamedArgumentNode("attr5", "value5"),
    ]

    args, kwargs, tags = parser.process_arguments()

    args, kwargs = set_names_and_defaults(
        args, kwargs, ["attr1", "attr2"], {"attr3": "value3"}
    )

    assert args == []
    assert kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value3",
        "attr5": "value5",
    }
    assert tags == []


def test_apply_prototype_clash_between_default_value_and_named_value():
    parser = ArgumentsParser(tokens=[])

    parser.nodes = [
        UnnamedArgumentNode("value1"),
        UnnamedArgumentNode("value2"),
        NamedArgumentNode("attr3", "value5"),
    ]

    args, kwargs, tags = parser.process_arguments()

    args, kwargs = set_names_and_defaults(
        args, kwargs, ["attr1", "attr2"], {"attr3": "value3"}
    )

    assert args == []
    assert kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value5",
    }
    assert tags == []


def test_apply_prototype_clash_between_positional_value_and_named_value():
    parser = ArgumentsParser(tokens=[])

    parser.nodes = [
        UnnamedArgumentNode("value1"),
        UnnamedArgumentNode("value2"),
        NamedArgumentNode("attr1", "value5"),
    ]

    args, kwargs, tags = parser.process_arguments()

    args, kwargs = set_names_and_defaults(
        args, kwargs, ["attr1", "attr2"], {"attr3": "value3"}
    )

    assert args == []
    assert kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value3",
    }
    assert tags == []


def test_apply_prototype_clash_in_prototype():
    parser = ArgumentsParser(tokens=[])

    parser.nodes = [
        UnnamedArgumentNode("value1"),
        UnnamedArgumentNode("value2"),
        NamedArgumentNode("attr3", "value3"),
    ]

    args, kwargs, tags = parser.process_arguments()

    args, kwargs = set_names_and_defaults(
        args, kwargs, ["attr1", "attr2"], {"attr1": "value7"}
    )

    assert args == []
    assert kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value3",
    }
    assert tags == []


def test_apply_prototype_positional_values_not_provided():
    parser = ArgumentsParser(tokens=[])

    parser.nodes = [UnnamedArgumentNode("value1")]

    args, kwargs, _ = parser.process_arguments()

    with pytest.raises(ValueError):
        set_names_and_defaults(args, kwargs, ["attr1", "attr2"], {"attr3": "value3"})


def test_apply_prototype_positional_values_without_name():
    parser = ArgumentsParser(tokens=[])

    parser.nodes = [UnnamedArgumentNode("value1"), UnnamedArgumentNode("value2")]

    args, kwargs, tags = parser.process_arguments()

    args, kwargs = set_names_and_defaults(args, kwargs, ["attr1"], {"attr3": "value3"})

    assert args == ["value2"]
    assert kwargs == {
        "attr1": "value1",
        "attr3": "value3",
    }
    assert tags == []


def test_apply_prototype_missing_positional_with_default():
    parser = ArgumentsParser(tokens=[])

    parser.nodes = [
        NamedArgumentNode("attr1", "value5"),
    ]

    args, kwargs, tags = parser.process_arguments()

    args, kwargs = set_names_and_defaults(args, kwargs, ["attr1"])

    assert args == []
    assert kwargs == {
        "attr1": "value5",
    }
    assert tags == []



from unittest.mock import Mock, patch

import pytest
from mau.lexers.base_lexer import BaseLexer, LexerError, TokenTypes
from mau.text_buffer.context import Context
from mau.text_buffer.text_buffer import TextBuffer
from mau.tokens.tokens import Token

from tests.helpers import dedent


def test_text_buffer_properties():
    mock_text_buffer = Mock()
    lex = BaseLexer(text_buffer=mock_text_buffer)

    assert lex._current_char == mock_text_buffer.current_char

    assert lex._current_line == mock_text_buffer.current_line

    assert lex._tail == mock_text_buffer.tail

    lex._nextline()
    mock_text_buffer.nextline.assert_called()

    lex._skip("four")
    mock_text_buffer.skip.assert_called_with(4)


def test_create_token():
    mock_text_buffer = Mock()
    lex = BaseLexer(mock_text_buffer)

    token = lex._create_token("sometype", "somevalue")
    assert token == Token("sometype", "somevalue")
    assert token.context == lex._text_buffer.context


@patch("mau.lexers.base_lexer.BaseLexer._nextline")
def test_create_tokens_from_line(mock_nextline):
    text_buffer = TextBuffer("Content")
    lex = BaseLexer(text_buffer)

    tokens = lex._create_tokens_from_line(TokenTypes.TEXT)

    assert tokens == [
        Token(TokenTypes.TEXT, "Content"),
        Token(TokenTypes.EOL),
    ]
    mock_nextline.assert_called()


def test_create_token_and_skip():
    mock_text_buffer = Mock()
    lex = BaseLexer(mock_text_buffer)
    lex._skip = Mock()

    token = lex._create_token_and_skip("sometype", "somevalue")
    assert token == Token("sometype", "somevalue")
    lex._skip.assert_called_with("somevalue")


def test_error():
    mock_text_buffer = Mock()
    lex = BaseLexer(mock_text_buffer)

    with pytest.raises(LexerError):
        lex._process_error()


def test_empty_text():
    mock_text_buffer = Mock()
    lex = BaseLexer(mock_text_buffer)
    lex.process()

    assert lex.tokens == []


def test_just_empty_lines():
    text_buffer = TextBuffer("\n")
    lex = BaseLexer(text_buffer)
    lex.process()

    assert lex.tokens == [Token(TokenTypes.EOL), Token(TokenTypes.EOL)]


def test_lines_with_only_spaces():
    text_buffer = TextBuffer("    \n    ")
    lex = BaseLexer(text_buffer)
    lex.process()

    assert lex.tokens == [Token(TokenTypes.EOL), Token(TokenTypes.EOL)]


def test_text():
    text = "Just simple text"
    text_buffer = TextBuffer(text)
    lex = BaseLexer(text_buffer)
    lex.process()

    assert lex.tokens == [Token(TokenTypes.TEXT, text), Token(TokenTypes.EOL)]


def test_multiple_lines():
    text = dedent(
        """
        This is text
        split into multiple lines

        with an empty line
        """
    )
    text_buffer = TextBuffer(text)
    lex = BaseLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "This is text"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.TEXT, "split into multiple lines"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOL),
        Token(TokenTypes.TEXT, "with an empty line"),
        Token(TokenTypes.EOL),
    ]


def test_positions_default_context():
    text = dedent(
        """
        This is a line of text

        ---

        This is another line of text
        """
    )
    text_buffer = TextBuffer(text)
    lex = BaseLexer(text_buffer)
    lex.process()

    assert [i.context.asdict() for i in lex.tokens] == [
        {"column": 0, "line": 0, "source": None, "text": "This is a line of text"},
        {"column": 22, "line": 0, "source": None, "text": "This is a line of text"},
        {"column": 0, "line": 1, "source": None, "text": ""},
        {"column": 0, "line": 2, "source": None, "text": "---"},
        {"column": 3, "line": 2, "source": None, "text": "---"},
        {"column": 0, "line": 3, "source": None, "text": ""},
        {
            "column": 0,
            "line": 4,
            "source": None,
            "text": "This is another line of text",
        },
        {
            "column": 28,
            "line": 4,
            "source": None,
            "text": "This is another line of text",
        },
    ]


def test_positions():
    text = dedent(
        """
        This is a line of text

        ---

        This is another line of text
        """
    )
    text_buffer = TextBuffer(text, Context(line=42, column=123, source="main"))
    lex = BaseLexer(text_buffer)
    lex.process()

    assert [i.context.asdict() for i in lex.tokens] == [
        {"column": 123, "line": 42, "source": "main", "text": "This is a line of text"},
        {"column": 145, "line": 42, "source": "main", "text": "This is a line of text"},
        {"column": 123, "line": 43, "source": "main", "text": ""},
        {"column": 123, "line": 44, "source": "main", "text": "---"},
        {"column": 126, "line": 44, "source": "main", "text": "---"},
        {"column": 123, "line": 45, "source": "main", "text": ""},
        {
            "column": 123,
            "line": 46,
            "source": "main",
            "text": "This is another line of text",
        },
        {
            "column": 151,
            "line": 46,
            "source": "main",
            "text": "This is another line of text",
        },
    ]



import hashlib


class Node:
    node_type = "node"

    @property
    def _content(self):
        return {"type": self.node_type}

    def accept(self, visitor, *args, **kwargs):
        method_name = f"_visit_{self.node_type}"

        try:
            method = getattr(visitor, method_name)
        except AttributeError:
            method = getattr(visitor, "_visit_default")

        return method(self, *args, **kwargs)

    def __eq__(self, other):
        try:
            return self._content == other._content
        except AttributeError:  # pragma: no cover
            return False

    def __repr__(self):  # pragma: no cover
        return f"{self._content}"

    def hash(self):  # pragma: no cover
        return hashlib.md5(str(self).encode("utf-8")).hexdigest()[:8]


class ValueNode(Node):
    node_type = "value_node"

    def __init__(self, value=None):
        self.value = value

    @property
    def _content(self):
        return {"type": self.node_type, "value": self.value}



import pytest
from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.page import BlockNode, ParagraphNode
from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_block_with_empty_body():
    source = """
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="default",
            content=[],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_parse_block_content():
    source = """
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="default",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("This is a paragraph."),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_parse_block_content_variables():
    source = """
    ----
    :answer:42
    The answer is {answer}.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="default",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("The answer is 42."),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_parse_block_content_external_variables():
    source = """
    :answer:42
    ----
    The answer is {answer}.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="default",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("The answer is 42."),
                        ]
                    )
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_block_without_closing_fences():
    parser = init_parser("----")

    with pytest.raises(ParserError):
        parser.parse()


def test_parse_block_secondary_content():
    source = """
    ----
    ----
    This is a paragraph that gets eaten.
    This is a second paragraph that gets eaten.

    This paragraph appears in the output.
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="default",
            content=[],
            secondary_content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode(
                                (
                                    "This is a paragraph that gets eaten. "
                                    "This is a second paragraph that gets eaten."
                                )
                            ),
                        ]
                    )
                ),
            ],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
        ParagraphNode(
            SentenceNode(
                [
                    TextNode("This paragraph appears in the output."),
                ]
            )
        ),
    ]


def test_parse_block_inside_block():
    source = """
    ----
    ++++
    ++++
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            blocktype="default",
            content=[
                BlockNode(
                    blocktype="default",
                    content=[],
                    secondary_content=[],
                    classes=[],
                    title=None,
                    engine="default",
                    preprocessor="none",
                    args=[],
                    kwargs={},
                )
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="default",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]



from mau.text_buffer.text_buffer import Context


def test_text_buffer_current_line():
    ctx = Context(line=1, column=7, source="main", text="sometext")

    assert ctx.asdict() == {
        "line": 1,
        "column": 7,
        "source": "main",
        "text": "sometext",
    }
    assert ctx == Context(line=1, column=7, source="main", text="sometext")
    assert ctx != (1, 7)



# This is a base class for parsers that collects common methods

from mau.lexers.base_lexer import BaseLexer, TokenTypes
from mau.parsers.arguments import set_names_and_defaults
from mau.text_buffer.text_buffer import TextBuffer
from mau.tokens.tokens import Token


class TokenError(ValueError):
    """
    This is an exception that parsers ca use to signal
    that the current function cannot process the upcoming
    tokens.
    This means that the parser as a context manager
    must reset the index, but also that the exception
    must not be propagated.
    """


class ParserError(ValueError):
    """This is a detailed parser error"""

    def __init__(self, token, message=None):
        super().__init__(message)

        self.token = token

    def __repr__(self):
        return f"{super().__str__()} - ({self.token.type}, {self.token.value})"

    def __str__(self):
        return self.__repr__()


class BaseParser:
    text_buffer_class = TextBuffer
    lexer_class = BaseLexer

    def __init__(self, tokens):
        # This is the position of the current token.
        self.index = -1

        # These are the tokens parsed by the parser.
        self.tokens = tokens

        # A stack for the parser's state.
        # Currently the state is represented only
        # by the current index in the input tokens.
        self._stack = []

        # These are the nodes created by the parsing.
        self.nodes = []

        # The last processed token. Used to detect loops.
        self.last_processed_token = Token(TokenTypes.EOF)

    @property
    def _current_token(self):
        """
        Returns the token being parsed.
        We often need to know which token we are currently
        parsing, but we might already have parsed all
        of them, so this convenience method wraps the
        possible index error.
        """
        if self.index < 0:
            return Token(TokenTypes.EOF)

        try:
            return self.tokens[self.index]
        except IndexError:
            return Token(TokenTypes.EOF)

    def _advance(self):
        if self.index != len(self.tokens):
            self.index += 1

    def _push(self):
        # Push the current state on the stack.
        self._stack.append(self.index)

    def _pop(self):
        # Get the state from the stack.
        return self._stack.pop()

    def __enter__(self):
        # The parser can be used as a context manager.
        # When we enter a new context we just need to
        # push the current state.
        self._push()

    def __exit__(self, etype, evalue, etrace):
        # This is automatically run when we leave the context.
        # The execution of a parser function can either return
        # None or raise an exception.
        # In the first case everything went well, and we can exit
        # the context manager after we restored the stack.
        # In the second case we need to check if the exception
        # is an expected one or not. If the exception is
        # a signal that the function failed we can suppress it.

        # First make sure we leave the stack as it was before
        # we entered the context.
        stacked_index = self._pop()

        # If there was no exception we can exit the
        # context manager and continue execution
        if etype is None:
            return True

        # If there was an exception we need to
        # actually backtrace and pretend we didn't
        # do anything. Be cautious and don't get caught.

        # Restore the original position
        self.index = stacked_index

        # If the exception is not among the managed
        # ones we need to signal that it has to
        # be propagated
        if etype not in [TokenError]:
            return False

        # At this point we know that there was
        # an exception but we can ignore it as
        # it is one of the expected ones
        return True

    def _save(self, node):
        # Store the node.
        self.nodes.append(node)

    def _check_token(
        self,
        token,
        ttype=None,
        tvalue=None,
        value_check_function=None,
    ):
        # This method performs a test on the current token,
        # figuring out if type and value correspond to those passed
        # as arguments. If type or value are not given they are not
        # tested.
        # If the test is successful the token is returned, otherwise
        # the TokenError exception is raised.
        # The argument value_check_function is a function that
        # can be passed to test the token value and shall return a boolean.

        check_type = ttype if ttype is not None else token.type
        if token.type != check_type:
            raise TokenError(f"Type of token {token} is not {check_type}")

        if tvalue is not None and token.value != tvalue:
            raise TokenError(f"Value of token {token} is not {tvalue}")

        if (
            value_check_function is not None
            and value_check_function(token.value) is False
        ):
            raise TokenError(f"Value of token {token} didn't pass check")

        return token

    def _process_functions(self):
        # The parse functions available in this parser
        return []

    def _error(self, message=None):
        raise ParserError(self._peek_token(), message)

    def _put_token(self, token):
        self.tokens.insert(self.index + 1, token)

    def _get_token(self, ttype=None, tvalue=None, check=None):
        """
        Return the next token and advances the index.
        This function returns the next token and then advances the index,
        and can optionally check its type or value (see _check_token).
        """

        self._advance()

        return self._check_token(self._current_token, ttype, tvalue, check)

    def _check_current_token(self, ttype, tvalue=None, check=None):
        """
        Just check the type and value of the current token without
        advancing the index.
        """

        return self._check_token(self._current_token, ttype, tvalue, check)

    def _peek_token(self, ttype=None, tvalue=None, check=None):
        """
        Return the next token without advancing the index.
        """

        try:
            token = self.tokens[self.index + 1]
        except IndexError:
            token = Token(TokenTypes.EOF)

        return self._check_token(token, ttype, tvalue, check)

    def _peek_token_is(self, ttype=None, tvalue=None, check=None):
        """
        Peek a token and check it.
        This works like peek_token, but returns a boolean
        instead of raising an exception.
        """
        try:
            self._peek_token(ttype, tvalue, check)
            return True
        except TokenError:
            return False

    def _force_token(self, ttype, tvalue=None):
        """
        Return the next token and advances the index,
        but forces the token to have a specific type
        and optionally a value.
        If the token doesn't match the provided values
        the function raises and error
        """
        try:
            token = self._get_token(ttype, tvalue)
        except TokenError:
            self._error(f"Expected token {Token(ttype, tvalue)}")

        return token

    def _collect(self, stop_tokens, preserve_escaped_stop_tokens=False):
        """
        Collect all tokens until one of the stop_tokens pops up.

        Escaped tokens (preceded by a literal "\\") are kept as they are.
        If a stop token is escaped it is included only if
        preserve_escaped_stop_tokens is set to True.
        """
        tokens = []

        # After all, at EOF the world ends.
        stop_tokens.append(Token(TokenTypes.EOF))

        while self._peek_token() not in stop_tokens:
            if self._peek_token() == Token(TokenTypes.LITERAL, "\\"):
                escape = self._get_token()

                if (
                    self._peek_token() not in stop_tokens
                    or preserve_escaped_stop_tokens
                ):
                    tokens.append(escape)

            tokens.append(self._get_token())

        return tokens

    def _collect_join(
        self,
        stop_tokens,
        join_with="",
        preserve_escaped_stop_tokens=False,
    ):
        """
        Collect all tokens until one of the stop_tokens
        pops up and join them in a single string.
        This works exactly like collect but returns
        a string of tokens joined with the given characters.
        """

        # Some tokens have value None, so this removes them
        token_values = [
            t.value
            for t in self._collect(stop_tokens, preserve_escaped_stop_tokens)
            if t.value != ""
        ]

        return join_with.join(token_values)

    # pylint: disable=inconsistent-return-statements
    def _set_names_and_defaults(
        self,
        args,
        kwargs,
        positional_names,
        default_values=None,
    ):  # pragma: no cover
        try:
            return set_names_and_defaults(
                args, kwargs, positional_names, default_values
            )
        except ValueError as exception:
            self._error(str(exception))

    def parse(self):
        """
        Run the parser on the lexed tokens.
        """

        # A loop on all lexed tokens until we reach EOF
        while not self._peek_token_is(TokenTypes.EOF):
            # This detects infinite loops created by incomplete
            # parsing functions. Those functions keep trying
            # to parse the same token, so if we spot that
            # we are doing it we should raise an error.
            next_token = self._peek_token()
            if next_token.match(self.last_processed_token):
                self._error("Loop detected, cannot parse token")  # pragma: no cover
            else:
                self.last_processed_token = next_token

            # Here we run all parsing functions provided by
            # the parser until one returns a sensible result
            result = False
            for process_function in self._process_functions():
                # The context manager wraps the function so that
                # any exception leaves the parsed tokens as they
                # were at the beginning of the function execution.
                #
                # If the parse function is successful it returns
                # True. If the function raises an exception the
                # variable result is not set and the value is False.
                # Any other result returned by the function
                # is ignored.
                with self:
                    result = process_function()

                if result is True:
                    # True means the function was successful
                    # and we can stop the loop.
                    break

            # If we get here and result is still False
            # we didn't find any function to parse the
            # current token.
            if result is False:
                self._error("Cannot parse token")

    @classmethod
    def analyse(cls, text, context, *args, **kwargs):
        text_buffer = cls.text_buffer_class(text, context)

        lex = cls.lexer_class(text_buffer)
        lex.process()

        par = cls(lex.tokens, *args, **kwargs)
        par.parse()

        return par



from typing import Optional

from mau.text_buffer.context import Context


class Token:
    """
    This represents a token.
    Tokens have a type, a value (the actual characters), and a context
    """

    def __init__(
        self, _type: str, value: Optional[str] = None, context: Optional[Context] = None
    ):
        self.type = _type
        self.context = context or Context()
        self.value = value or ""

    def __repr__(self):
        return f'Token({self.type}, "{self.value}", {self.context})'

    def __eq__(self, other):
        try:
            if self.value == "" or other.value == "":
                return self.type == other.type

            return (self.type, self.value) == (
                other.type,
                other.value,
            )
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self.type, self.value))

    def __len__(self):
        if self.value:
            return len(self.value)

        return 0

    def __bool__(self):
        return True

    def match(self, other):
        # Match is different from __eq__ because it also
        # checks the context.
        return self == other and self.context == other.context



from mau.lexers.arguments_lexer import ArgumentsLexer
from mau.lexers.base_lexer import TokenTypes
from mau.text_buffer.context import Context
from mau.text_buffer.text_buffer import TextBuffer
from mau.tokens.tokens import Token


def test_single_unnamed_argument():
    text_buffer = TextBuffer("value1")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.EOL),
    ]


def test_single_named_argument():
    text_buffer = TextBuffer("argument1=value1")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "argument1"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.EOL),
    ]


def test_multiple_unnamed_arguments():
    text_buffer = TextBuffer("value1, value2")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.LITERAL, ","),
        Token(TokenTypes.WHITESPACE, " "),
        Token(TokenTypes.TEXT, "value2"),
        Token(TokenTypes.EOL),
    ]


def test_multiple_named_arguments():
    text_buffer = TextBuffer("argument1=value1, argument2=value2")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "argument1"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.LITERAL, ","),
        Token(TokenTypes.WHITESPACE, " "),
        Token(TokenTypes.TEXT, "argument2"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.TEXT, "value2"),
        Token(TokenTypes.EOL),
    ]


def test_mixed_arguments():
    text_buffer = TextBuffer("value1, value2,argument1=value1, argument2=value2")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.LITERAL, ","),
        Token(TokenTypes.WHITESPACE, " "),
        Token(TokenTypes.TEXT, "value2"),
        Token(TokenTypes.LITERAL, ","),
        Token(TokenTypes.TEXT, "argument1"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.LITERAL, ","),
        Token(TokenTypes.WHITESPACE, " "),
        Token(TokenTypes.TEXT, "argument2"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.TEXT, "value2"),
        Token(TokenTypes.EOL),
    ]


def test_quotes():
    text_buffer = TextBuffer('argument1="value1,value2"')
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "argument1"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.LITERAL, ","),
        Token(TokenTypes.TEXT, "value2"),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.EOL),
    ]


def test_spaces():
    text_buffer = TextBuffer("argument1=value1 value2")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "argument1"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.WHITESPACE, " "),
        Token(TokenTypes.TEXT, "value2"),
        Token(TokenTypes.EOL),
    ]


def test_escaped_quotes():
    text_buffer = TextBuffer(r"Argument \"with\" quotes")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "Argument"),
        Token(TokenTypes.WHITESPACE, " "),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.TEXT, "with"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.WHITESPACE, " "),
        Token(TokenTypes.TEXT, "quotes"),
        Token(TokenTypes.EOL),
    ]


def test_context():
    text_buffer = TextBuffer("argument1=value1")
    lex = ArgumentsLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "argument1"),
        Token(TokenTypes.LITERAL, "="),
        Token(TokenTypes.TEXT, "value1"),
        Token(TokenTypes.EOL),
    ]

    assert [t.context for t in lex.tokens] == [
        Context(line=0, column=0, source=None, text="argument1=value1"),
        Context(line=0, column=9, source=None, text="argument1=value1"),
        Context(line=0, column=10, source=None, text="argument1=value1"),
        Context(line=0, column=16, source=None, text="argument1=value1"),
    ]



from mau.lexers.base_lexer import TokenTypes
from mau.lexers.preprocess_variables_lexer import PreprocessVariablesLexer
from mau.nodes.inline import TextNode
from mau.parsers.base_parser import BaseParser
from mau.parsers.environment import Environment
from mau.tokens.tokens import Token


class PreprocessVariablesParser(BaseParser):
    lexer_class = PreprocessVariablesLexer

    def __init__(self, tokens, environment=None):
        super().__init__(tokens)

        self.environment = environment or Environment()

    def _process_verbatim(self):
        self._get_token(TokenTypes.LITERAL, "`")
        text = self._collect_join(
            [Token(TokenTypes.LITERAL, "`")],
            preserve_escaped_stop_tokens=True,
        )
        self._get_token(TokenTypes.LITERAL, "`")

        text = f"`{text}`"

        self._save(TextNode(text))

        return True

    def _process_escaped_char(self):
        self._get_token(TokenTypes.LITERAL, "\\")

        char = self._get_token().value

        if char not in "{}":
            char = f"\\{char}"

        self._save(TextNode(char))

        return True

    def _process_curly(self):
        self._get_token(TokenTypes.LITERAL, "{")
        variable_name = self._collect_join(stop_tokens=[Token(TokenTypes.LITERAL, "}")])
        self._get_token(TokenTypes.LITERAL, "}")

        try:
            variable_value = self.environment.getvar(variable_name)

            self._save(TextNode(variable_value))
        except KeyError:
            self._error(f'Attribute "{variable_name}" has not been defined')

        return True

    def _process_pass(self):
        self._save(TextNode(self._get_token().value))

        return True

    def _process_functions(self):
        return [
            self._process_escaped_char,
            self._process_verbatim,
            self._process_curly,
            self._process_pass,
        ]

    def parse(self):
        super().parse()

        # After having parsed the text and replaced the
        # variables, this should return a piece of text again
        text = "".join([str(i.value) for i in self.nodes])
        self.nodes = [TextNode(text)]



from mau.nodes.footnotes import CommandFootnotesNode, FootnoteNode, FootnotesEntryNode
from mau.nodes.inline import TextNode, VerbatimNode


def test_footnote_node():
    node = FootnoteNode([VerbatimNode("somevalue"), TextNode("othervalue")])

    assert node.content == [VerbatimNode("somevalue"), TextNode("othervalue")]
    assert node.number is None
    assert node.reference_anchor is None
    assert node.content_anchor is None
    assert node.node_type == "footnote"
    assert node == FootnoteNode([VerbatimNode("somevalue"), TextNode("othervalue")])


def test_footnote_node_with_number_and_anchor():
    node = FootnoteNode(
        [VerbatimNode("somevalue"), TextNode("othervalue")],
        number="3",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    assert node.content == [VerbatimNode("somevalue"), TextNode("othervalue")]
    assert node.number == "3"
    assert node.reference_anchor == "someanchor"
    assert node.content_anchor == "someanchor-def"
    assert node.node_type == "footnote"
    assert node == FootnoteNode(
        [VerbatimNode("somevalue"), TextNode("othervalue")],
        number="3",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )


def test_footnotes_entry_node():
    node = FootnotesEntryNode(
        content=TextNode("Some text"),
        number="1",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    assert node.content == TextNode("Some text")
    assert node.number == "1"
    assert node.reference_anchor == "someanchor"
    assert node.content_anchor == "someanchor-def"
    assert node.node_type == "footnotes_entry"
    assert node == FootnotesEntryNode(
        content=TextNode("Some text"),
        number="1",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )


def test_command_footnotes_node():
    node = CommandFootnotesNode(
        entries=[TextNode("somevalue1"), TextNode("somevalue2")],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.entries == [TextNode("somevalue1"), TextNode("somevalue2")]
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "command_footnotes"
    assert node == CommandFootnotesNode(
        entries=[TextNode("somevalue1"), TextNode("somevalue2")],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )



from mau.lexers.base_lexer import TokenTypes
from mau.lexers.text_lexer import TextLexer
from mau.text_buffer.text_buffer import TextBuffer
from mau.tokens.tokens import Token


def test_normal_text():
    text_buffer = TextBuffer("Normal text")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "Normal"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "text"),
        Token(TokenTypes.EOL),
    ]


def test_underscore():
    text_buffer = TextBuffer("_underscore_")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.TEXT, "underscore"),
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.EOL),
    ]


def test_star():
    text_buffer = TextBuffer("*star*")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "*"),
        Token(TokenTypes.TEXT, "star"),
        Token(TokenTypes.LITERAL, "*"),
        Token(TokenTypes.EOL),
    ]


def test_caret():
    text_buffer = TextBuffer("^caret^")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "^"),
        Token(TokenTypes.TEXT, "caret"),
        Token(TokenTypes.LITERAL, "^"),
        Token(TokenTypes.EOL),
    ]


def test_tilde():
    text_buffer = TextBuffer("~tilde~")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "~"),
        Token(TokenTypes.TEXT, "tilde"),
        Token(TokenTypes.LITERAL, "~"),
        Token(TokenTypes.EOL),
    ]


def test_backtick():
    text_buffer = TextBuffer("`backtick`")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.TEXT, "backtick"),
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.EOL),
    ]


def test_curly_braces():
    text_buffer = TextBuffer("{curly}")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "{"),
        Token(TokenTypes.TEXT, "curly"),
        Token(TokenTypes.LITERAL, "}"),
        Token(TokenTypes.EOL),
    ]


def test_round_brackets():
    text_buffer = TextBuffer("(round)")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "round"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
    ]


def test_square_brackets():
    text_buffer = TextBuffer("[square]")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "square"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.EOL),
    ]


def test_macro():
    text_buffer = TextBuffer("[macro](value1,value2)")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "macro"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "value1,value2"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
    ]


def test_macro_named_attributes():
    text_buffer = TextBuffer("[macro](attr1=value1,attr2=value2)")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "macro"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "attr1=value1,attr2=value2"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
    ]


def test_escaped_underscore():
    text_buffer = TextBuffer(r"\_underscore\_")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.TEXT, "underscore"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.EOL),
    ]


def test_escaped_square_brackets():
    text_buffer = TextBuffer(r"\[square\]")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "square"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.EOL),
    ]


def test_escaped_round_brackets():
    text_buffer = TextBuffer(r"\(round\)")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "round"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
    ]


def test_escaped_curly_braces():
    text_buffer = TextBuffer(r"\{curly\}")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "{"),
        Token(TokenTypes.TEXT, "curly"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "}"),
        Token(TokenTypes.EOL),
    ]


def test_escaped_quotes():
    text_buffer = TextBuffer(r"\"quotes\"")
    lex = TextLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.TEXT, "quotes"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.EOL),
    ]



from mau.lexers.arguments_lexer import ArgumentsLexer
from mau.lexers.base_lexer import TokenTypes
from mau.nodes.arguments import NamedArgumentNode, UnnamedArgumentNode
from mau.parsers.base_parser import BaseParser
from mau.tokens.tokens import Token


class ArgumentsParser(BaseParser):
    lexer_class = ArgumentsLexer

    def __init__(self, tokens):
        super().__init__(tokens)

        # This flag is turned on as soon as
        # a named argument is parsed
        self._named_arguments = False

    def _process_functions(self):
        return [
            self._process_eol,
            self._process_named_argument,
            self._process_unnamed_argument,
        ]

    def _process_eol(self):
        # This simply parses the end of line, ignoring it

        self._get_token(TokenTypes.EOL)

        return True

    def _process_named_argument(self):
        # This parses a named argument

        key = self._get_token(TokenTypes.TEXT).value
        self._get_token(TokenTypes.LITERAL, "=")

        # Values can be surrounded by quotes
        if self._peek_token_is(TokenTypes.LITERAL, '"'):
            self._get_token(TokenTypes.LITERAL, '"')
            value = self._collect_join([Token(TokenTypes.LITERAL, '"')])
            self._get_token(TokenTypes.LITERAL, '"')
        else:
            value = self._collect_join([Token(TokenTypes.LITERAL, ",")])

        # The comma is not there after the last argument
        with self:
            self._get_token(TokenTypes.LITERAL, ",")

        # Leading spaces are removed
        with self:
            self._get_token(TokenTypes.WHITESPACE)

        # Leading and trailing spaces are removed
        self._save(NamedArgumentNode(key, value.strip()))

        # Mark the beginning of named arguments
        self._named_arguments = True

        return True

    def _process_unnamed_argument(self):
        # This parses an unnamed argument

        if self._named_arguments:
            self._error("Unnamed arguments after named arguments are forbidden")

        # Values can be surrounded by quotes
        if self._peek_token_is(TokenTypes.LITERAL, '"'):
            self._get_token(TokenTypes.LITERAL, '"')
            value = self._collect_join([Token(TokenTypes.LITERAL, '"')])
            self._get_token(TokenTypes.LITERAL, '"')
        else:
            value = self._collect_join([Token(TokenTypes.LITERAL, ",")])

        # The comma is not there after the last argument
        with self:
            self._get_token(TokenTypes.LITERAL, ",")

        # Leading spaces are removed
        with self:
            self._get_token(TokenTypes.WHITESPACE)

        self._save(UnnamedArgumentNode(value))

        return True

    def process_arguments(self):
        """
        Process the extracted nodes and converts them into args (unnamed arguments),
        kwargs (named arguments), and tags (unnamed arguments whose name begins
        with "#".
        """

        args = [
            node.value for node in self.nodes if node.node_type == "unnamed_argument"
        ]

        kwargs = {
            node.key: node.value
            for node in self.nodes
            if node.node_type == "named_argument"
        }

        # Isolate tags
        tags = [i for i in args if i.startswith("#")]

        # Isolate non-tags
        args = [i for i in args if i not in tags]

        # Remove the "#" from tags
        tags = [i[1:] for i in tags]

        return args, kwargs, tags



import pytest
from mau.lexers.main_lexer import MainLexer
from mau.nodes.footnotes import CommandFootnotesNode
from mau.nodes.inline import LinkNode, SentenceNode, StyleNode, TextNode
from mau.nodes.page import CommandTocNode, HorizontalRuleNode, ParagraphNode
from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

# from mau.parsers.base_parser import ParserError


init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_discards_empty_lines():
    source = "\n\n\n\n"

    assert runner(source).nodes == []


def test_horizontal_rule():
    source = "---"

    expected = [HorizontalRuleNode()]

    assert runner(source).nodes == expected


def test_horizontal_rule_with_arguments():
    source = """
    [break,arg1=value1]
    ---
    """

    expected = [
        HorizontalRuleNode(args=["break"], kwargs={"arg1": "value1"}),
    ]

    assert runner(source).nodes == expected


def test_parse_single_line_comments():
    source = "// Just a comment"

    assert runner(source).nodes == []


def test_parse_multi_line_comments():
    source = """
    ////
    This is a
    multiline
    comment
    ////
    """

    assert runner(source).nodes == []


def test_parse_multi_line_comments_with_dangerous_characters():
    source = """
    ////
    .This is a
    [multiline]
    ----
    comment
    ////
    """

    assert runner(source).nodes == []


def test_parse_open_multi_line_comments():
    source = """
    ////
    .This is a
    [multiline]
    ----
    comment
    """

    with pytest.raises(ParserError):
        runner(source)


def test_parse_paragraphs():
    source = """
    This is a paragraph.
    This is part of the same paragraph.

    This is a new paragraph.
    """

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    TextNode(
                        "This is a paragraph. This is part of the same paragraph."
                    ),
                ]
            ),
            args=[],
            kwargs={},
        ),
        ParagraphNode(
            SentenceNode(
                [
                    TextNode("This is a new paragraph."),
                ]
            ),
            args=[],
            kwargs={},
        ),
    ]


def test_parse_paragraph_starting_with_a_macro():
    source = "[link](http://some.where,This) is the link I want"

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    LinkNode(target="http://some.where", text="This"),
                    TextNode(" is the link I want"),
                ]
            )
        )
    ]


def test_attributes():
    source = "[value1,someattr1=somevalue1,someattr2=somevalue2]"

    parser = init_parser(source)
    parser.parse()

    assert parser.arguments == (
        ["value1"],
        {
            "someattr1": "somevalue1",
            "someattr2": "somevalue2",
        },
        [],
    )


def test_attributes_paragraph():
    source = """
    [value1,someattr1=somevalue1,someattr2=somevalue2]
    This is text
    """

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [TextNode("This is text")],
            ),
            args=["value1"],
            kwargs={"someattr1": "somevalue1", "someattr2": "somevalue2"},
        ),
    ]


def test_attributes_with_variables():
    source = """
    :attrs:value1,someattr1=somevalue1,someattr2=somevalue2

    [{attrs}]
    This is text
    """

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [TextNode("This is text")],
            ),
            args=["value1"],
            kwargs={"someattr1": "somevalue1", "someattr2": "somevalue2"},
        ),
    ]


def test_command():
    source = "::somecommand:arg1, arg2, name1=value1, name2=value2"

    assert runner(source).nodes == []


def test_command_without_arguments():
    source = "::somecommand:"

    assert runner(source).nodes == []


def test_command_toc():
    source = "::toc:arg1, arg2, #tag1, name1=value1, name2=value2"

    assert runner(source).nodes == [
        CommandTocNode(
            [],
            args=["arg1", "arg2"],
            kwargs={"name1": "value1", "name2": "value2"},
            tags=["tag1"],
        ),
    ]


def test_command_footnotes():
    source = "::footnotes:arg1, arg2, #tag1, name1=value1, name2=value2"

    parser = init_parser(source)
    parser.footnote_entries = []
    parser.parse()

    assert parser.nodes == [
        CommandFootnotesNode(
            parser.footnote_entries,
            args=["arg1", "arg2"],
            kwargs={"name1": "value1", "name2": "value2"},
            tags=["tag1"],
        ),
    ]


def test_style_underscore():
    source = """
    This is _underscore_ text
    """

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    TextNode("This is "),
                    StyleNode(
                        "underscore",
                        SentenceNode(
                            [
                                TextNode("underscore"),
                            ],
                        ),
                    ),
                    TextNode(" text"),
                ],
            )
        )
    ]


def test_style_at_beginning():
    source = """
    *This is star text*
    """

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    StyleNode(
                        "star",
                        SentenceNode(
                            [
                                TextNode("This is star text"),
                            ]
                        ),
                    ),
                ],
            )
        )
    ]


def test_style_not_closed():
    source = r"""
    This ` is a backtick and this _an underscore
    """

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    TextNode("This ` is a backtick and this _an underscore"),
                ],
            )
        )
    ]


def test_style_escape_markers():
    source = r"""
    This is \_underscore\_ and this is \`verbatim\`
    """

    assert runner(source).nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    TextNode("This is _underscore_ and this is `verbatim`"),
                ],
            )
        )
    ]



from mau.lexers.base_lexer import TokenTypes
from mau.lexers.preprocess_variables_lexer import PreprocessVariablesLexer
from mau.text_buffer.text_buffer import TextBuffer
from mau.tokens.tokens import Token


def test_normal_text():
    text_buffer = TextBuffer("Some text")
    lex = PreprocessVariablesLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "S"),
        Token(TokenTypes.TEXT, "o"),
        Token(TokenTypes.TEXT, "m"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "x"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.EOL),
    ]


def test_match_only_backticks_and_curly_braces():
    text_buffer = TextBuffer("Normal text `{curly}` _other_ *text*")
    lex = PreprocessVariablesLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "N"),
        Token(TokenTypes.TEXT, "o"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.TEXT, "m"),
        Token(TokenTypes.TEXT, "a"),
        Token(TokenTypes.TEXT, "l"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "x"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.LITERAL, "{"),
        Token(TokenTypes.TEXT, "c"),
        Token(TokenTypes.TEXT, "u"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.TEXT, "l"),
        Token(TokenTypes.TEXT, "y"),
        Token(TokenTypes.LITERAL, "}"),
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "_"),
        Token(TokenTypes.TEXT, "o"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "h"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.TEXT, "_"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "*"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "x"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "*"),
        Token(TokenTypes.EOL),
    ]


def test_escape_curly_braces():
    text_buffer = TextBuffer(r"Normal text \{curly\} _other_ *text*")
    lex = PreprocessVariablesLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "N"),
        Token(TokenTypes.TEXT, "o"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.TEXT, "m"),
        Token(TokenTypes.TEXT, "a"),
        Token(TokenTypes.TEXT, "l"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "x"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "{"),
        Token(TokenTypes.TEXT, "c"),
        Token(TokenTypes.TEXT, "u"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.TEXT, "l"),
        Token(TokenTypes.TEXT, "y"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "}"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "_"),
        Token(TokenTypes.TEXT, "o"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "h"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.TEXT, "_"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "*"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "x"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "*"),
        Token(TokenTypes.EOL),
    ]


def test_preserve_escapes():
    text_buffer = TextBuffer(r"Normal \text \_other\_")
    lex = PreprocessVariablesLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "N"),
        Token(TokenTypes.TEXT, "o"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.TEXT, "m"),
        Token(TokenTypes.TEXT, "a"),
        Token(TokenTypes.TEXT, "l"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "x"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.TEXT, "_"),
        Token(TokenTypes.TEXT, "o"),
        Token(TokenTypes.TEXT, "t"),
        Token(TokenTypes.TEXT, "h"),
        Token(TokenTypes.TEXT, "e"),
        Token(TokenTypes.TEXT, "r"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.TEXT, "_"),
        Token(TokenTypes.EOL),
    ]



from unittest.mock import patch

from mau.lexers.main_lexer import MainLexer
from mau.nodes.page import HeaderNode
from mau.parsers.environment import Environment
from mau.parsers.main_parser import MainParser, header_anchor

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


@patch("mau.parsers.main_parser.hashlib.md5")
def test_default_header_anchor_function(mock_md5):
    mock_md5().hexdigest.return_value = "XXYY"

    assert header_anchor("Some Words 1234 56", "1") == "some-words-1234-56-XXYY"


@patch("mau.parsers.main_parser.hashlib.md5")
def test_default_header_anchor_function_multiple_spaces(mock_md5):
    mock_md5().hexdigest.return_value = "XXYY"

    assert (
        header_anchor("Some    Words     1234    56", "1") == "some-words-1234-56-XXYY"
    )


@patch("mau.parsers.main_parser.hashlib.md5")
def test_default_header_anchor_function_filter_characters(mock_md5):
    mock_md5().hexdigest.return_value = "XXYY"

    assert header_anchor("Some #Words @ 12!34 56", "1") == "some-words-1234-56-XXYY"


def test_custom_header_anchor_function():
    source = """
    = Title of the section
    """

    config = {"header_anchor_function": lambda text, level: "XXXXXY"}
    env_runner = parser_runner_factory(
        MainLexer, MainParser, environment=Environment(config, "mau")
    )

    assert env_runner(source).nodes == [
        HeaderNode("Title of the section", "1", "XXXXXY")
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_level_1(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    = Title of the section
    """

    assert runner(source).nodes == [HeaderNode("Title of the section", "1", "XXXXXX")]


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_level_3(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    === Title of a subsection
    """

    assert runner(source).nodes == [HeaderNode("Title of a subsection", "3", "XXXXXX")]


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_collect_headers(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    = Header 1
    == Header 1.1
    == Header 1.2
    = Header 2
    == Header 2.1
    === Header 2.1.1
    """

    parser = runner(source)

    assert parser.nodes == [
        HeaderNode("Header 1", "1", "Header 1-XXXXXX"),
        HeaderNode("Header 1.1", "2", "Header 1.1-XXXXXX"),
        HeaderNode("Header 1.2", "2", "Header 1.2-XXXXXX"),
        HeaderNode("Header 2", "1", "Header 2-XXXXXX"),
        HeaderNode("Header 2.1", "2", "Header 2.1-XXXXXX"),
        HeaderNode("Header 2.1.1", "3", "Header 2.1.1-XXXXXX"),
    ]

    assert parser.headers == [
        HeaderNode("Header 1", "1", "Header 1-XXXXXX"),
        HeaderNode("Header 1.1", "2", "Header 1.1-XXXXXX"),
        HeaderNode("Header 1.2", "2", "Header 1.2-XXXXXX"),
        HeaderNode("Header 2", "1", "Header 2-XXXXXX"),
        HeaderNode("Header 2.1", "2", "Header 2.1-XXXXXX"),
        HeaderNode("Header 2.1.1", "3", "Header 2.1.1-XXXXXX"),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_attributes_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [value1,someattr1=somevalue1,someattr2=somevalue2]
    = Header
    """

    assert runner(source).nodes == [
        HeaderNode(
            "Header",
            "1",
            "Header-XXXXXX",
            args=["value1"],
            kwargs={"someattr1": "somevalue1", "someattr2": "somevalue2"},
        ),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_single_tag_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [value1, #tag1, value2,someattr1=somevalue1,someattr2=somevalue2]
    = Header
    """

    assert runner(source).nodes == [
        HeaderNode(
            "Header",
            "1",
            "Header-XXXXXX",
            args=["value1", "value2"],
            kwargs={"someattr1": "somevalue1", "someattr2": "somevalue2"},
            tags=["tag1"],
        ),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_multiple_tags_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [value1, #tag1, value2,#tag2,someattr1=somevalue1,someattr2=somevalue2]
    = Header
    """

    assert runner(source).nodes == [
        HeaderNode(
            "Header",
            "1",
            "Header-XXXXXX",
            args=["value1", "value2"],
            kwargs={"someattr1": "somevalue1", "someattr2": "somevalue2"},
            tags=["tag1", "tag2"],
        ),
    ]



from mau.nodes.inline import (
    TextNode,
    SentenceNode,
    ListItemNode,
)

from mau.nodes.source import SourceNode, CalloutNode, CalloutsEntryNode
from mau.nodes.page import (
    HorizontalRuleNode,
    ParagraphNode,
    HeaderNode,
    ListNode,
    ContentNode,
    ContentImageNode,
    BlockNode,
    CommandTocNode,
)

from mau.visitors.jinja_visitor import JinjaVisitor
from mau.nodes.footnotes import FootnotesEntryNode, CommandFootnotesNode
from mau.nodes.references import ReferencesEntryNode, CommandReferencesNode


def test_page_horizontal_rule_node():
    templates = {
        "text.j2": "{{ value }}",
        "horizontal_rule.j2": (
            "--- - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = HorizontalRuleNode(args=args, kwargs=kwargs, tags=tags)

    result = visitor.visit(node)

    assert result == "--- - arg1,arg2 - key1:value1 - tag1,tag2"


def test_page_paragraph_node():
    templates = {
        "text.j2": "{{ value }}",
        "paragraph.j2": (
            "{{ content }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = ParagraphNode(
        TextNode("Just some text"), args=args, kwargs=kwargs, tags=tags
    )

    result = visitor.visit(node)

    assert result == "Just some text - arg1,arg2 - key1:value1 - tag1,tag2"


def test_page_header_node():
    templates = {
        "text.j2": "{{ value }}",
        "header.j2": (
            "{{ value }} - {{ level }} - {{ anchor }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = HeaderNode(
        "Just some text", "3", "someanchor", args=args, kwargs=kwargs, tags=tags
    )

    result = visitor.visit(node)

    assert (
        result
        == "Just some text - 3 - someanchor - arg1,arg2 - key1:value1 - tag1,tag2"
    )


def test_page_list_node():
    templates = {
        "text.j2": "{{ value }}",
        "list_item.j2": "{{ level }}:{{ content }}",
        "list.j2": (
            "{{ ordered }} - {{ items }} - {{ main_node }} - {{ args | join(',') }} - "
            "{{ kwargs|items|map('join', '=')|join(',') }} - {{ tags | join(',') }} - "
            "{{ kwargs.start }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1", "start": 4}
    tags = ["tag1", "tag2"]
    node = ListNode(
        True,
        [ListItemNode("4", TextNode("Just some text."))],
        True,
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert (
        result
        == "True - 4:Just some text. - True - arg1,arg2 - key1=value1,start=4 - tag1,tag2 - 4"
    )


def test_page_content_node():
    templates = {
        "text.j2": "{{ value }}",
        "content-sometype.j2": (
            "{{ title }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = ContentNode(
        "sometype", TextNode("sometitle"), args=args, kwargs=kwargs, tags=tags
    )

    result = visitor.visit(node)

    assert result == "sometitle - arg1,arg2 - key1:value1 - tag1,tag2"


def test_content_image_node():
    templates = {
        "text.j2": "{{ value }}",
        "content_image.j2": (
            "{{ uri }} - {{ alt_text }} - {{ classes | join(',') }} - "
            "{{ title }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = ContentImageNode(
        "someuri",
        "sometext",
        ["class1", "class2"],
        TextNode("sometitle"),
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert (
        result
        == "someuri - sometext - class1,class2 - sometitle - arg1,arg2 - key1:value1 - tag1,tag2"
    )


def test_page_block_node_standard_block_template():
    templates = {
        "text.j2": "{{ value }}",
        "block.j2": (
            "{{ blocktype }} - {{ content }} - {{ secondary_content }} - "
            "{{ classes | join(',') }} - {{ title }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        blocktype="someblock",
        content=[TextNode("my content")],
        secondary_content=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=TextNode("sometitle"),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == (
        "someblock - my content - my secondary content - class1,class2 - "
        "sometitle - arg1,arg2 - key1:value1 - tag1,tag2"
    )


def test_page_block_node_blocktype_template_has_precedence():
    templates = {
        "text.j2": "{{ value }}",
        "block-someblock.j2": "The blocktype template",
        "block.j2": "The wrong template",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        blocktype="someblock",
        content=[TextNode("my content")],
        secondary_content=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=TextNode("sometitle"),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "The blocktype template"


def test_page_block_node_engine_template_has_precedence():
    templates = {
        "text.j2": "{{ value }}",
        "block-someengine.j2": "The engine template",
        "block-someblock.j2": "The wrong template",
        "block.j2": "The wrong template",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        blocktype="someblock",
        content=[TextNode("my content")],
        secondary_content=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=TextNode("sometitle"),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "The engine template"


def test_page_block_node_engine_and_blocktype_template_has_precedence():
    templates = {
        "text.j2": "{{ value }}",
        "block-someengine-someblock.j2": "The engine+block template",
        "block-someengine.j2": "The wrong template",
        "block-someblock.j2": "The wrong template",
        "block.j2": "The wrong template",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        blocktype="someblock",
        content=[TextNode("my content")],
        secondary_content=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=TextNode("sometitle"),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "The engine+block template"


def test_page_command_toc_node():
    templates = {
        "text.j2": "{{ value }}",
        "toc.j2": (
            "{{ entries }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
        "toc_entry.j2": (
            "{{ anchor }}:{{ value }}{% if children %} - " "{{ children }}{% endif %}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = CommandTocNode(
        entries=[
            HeaderNode("Header 1", "1", "header-1"),
            HeaderNode("Header 1.1", "2", "header-1-1"),
            HeaderNode("Header 2", "1", "header-2"),
        ],
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == (
        "header-1:Header 1 - header-1-1:Header 1.1header-2:Header 2 - "
        "arg1,arg2 - key1:value1 - tag1,tag2"
    )


def test_page_command_toc_node_exclude():
    templates = {
        "text.j2": "{{ value }}",
        "toc.j2": (
            "{{ entries }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
        "toc_entry.j2": (
            "{{ anchor }}:{{ value }}{% if children %} - " "{{ children }}{% endif %}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"exclude_tag": "notoc"}
    tags = ["tag1", "tag2"]
    node = CommandTocNode(
        entries=[
            HeaderNode("Header 1", "1", "header-1", tags=["notoc"]),
            HeaderNode("Header 1.1", "2", "header-1-1"),
            HeaderNode("Header 2", "1", "header-2"),
        ],
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "header-2:Header 2 - arg1,arg2 - exclude_tag:notoc - tag1,tag2"


def test_page_command_footnotes_node():
    templates = {
        "text.j2": "{{ value }}",
        "footnotes.j2": "{{ entries }}",
        "footnotes_entry.j2": (
            "{{ content }}:{{ number }}:" "{{ content_anchor }}:{{ reference_anchor }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = CommandFootnotesNode(
        entries=[
            FootnotesEntryNode(
                [TextNode("Footnote 1")], "1", "anchor-1", "anchor-1-def"
            ),
            FootnotesEntryNode(
                [TextNode("Footnote 2")], "2", "anchor-2", "anchor-2-def"
            ),
        ],
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert (
        result == "Footnote 1:1:anchor-1-def:anchor-1Footnote 2:2:anchor-2-def:anchor-2"
    )


def test_page_command_references_single_node():
    templates = {
        "sentence.j2": "{{ content }}",
        "paragraph.j2": "{{ content }}",
        "text.j2": "{{ value }}",
        "references.j2": "{{ entries }}",
        "references_entry-content_type1.j2": (
            "{{ content_type }}:{{ name }}:{{ content }}:"
            "{{ number }}:{{ content_anchor }}:{{ reference_anchor }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = CommandReferencesNode(
        content_type="content_type1",
        name="value1",
        entries={
            ("content_type1", "value1"): ReferencesEntryNode(
                "content_type1",
                "value1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 1"),
                            ]
                        )
                    ),
                ],
                number=1,
                reference_anchor="ref-content_type1-1-XXYY",
                content_anchor="cnt-content_type1-1-XXYY",
            ),
            ("content_type1", "value2"): ReferencesEntryNode(
                "content_type1",
                "value2",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 2"),
                            ]
                        )
                    ),
                ],
                number=2,
                reference_anchor="ref-content_type1-2-XXYY",
                content_anchor="cnt-content_type1-2-XXYY",
            ),
            ("content_type2", "value1"): ReferencesEntryNode(
                "content_type2",
                "value1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 2, value 1"),
                            ]
                        )
                    ),
                ],
                number=1,
                reference_anchor="ref-content_type2-1-XXYY",
                content_anchor="cnt-content_type2-1-XXYY",
            ),
        },
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == (
        "content_type1:value1:Content type 1, value 1:1:"
        "cnt-content_type1-1-XXYY:ref-content_type1-1-XXYY"
    )


def test_page_command_references_multiple_nodes():
    templates = {
        "sentence.j2": "{{ content }}",
        "paragraph.j2": "{{ content }}",
        "text.j2": "{{ value }}",
        "references.j2": "{{ entries }}",
        "references_entry-content_type1.j2": (
            "{{ content_type }}:{{ name }}:{{ content }}:{{ number }}:"
            "{{ title }}:{{ content_anchor }}:{{ reference_anchor }}::"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = CommandReferencesNode(
        content_type="content_type1",
        name=None,
        entries={
            ("content_type1", "value1"): ReferencesEntryNode(
                "content_type1",
                "value1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 1"),
                            ]
                        )
                    ),
                ],
                number=1,
                title=SentenceNode([TextNode("Some title 1.1")]),
                reference_anchor="ref-content_type1-1-XXYY",
                content_anchor="cnt-content_type1-1-XXYY",
            ),
            ("content_type1", "value2"): ReferencesEntryNode(
                "content_type1",
                "value2",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 1, value 2"),
                            ]
                        )
                    ),
                ],
                number=2,
                title=SentenceNode([TextNode("Some title 1.2")]),
                reference_anchor="ref-content_type1-2-XXYY",
                content_anchor="cnt-content_type1-2-XXYY",
            ),
            ("content_type2", "value1"): ReferencesEntryNode(
                "content_type2",
                "value1",
                content=[
                    ParagraphNode(
                        SentenceNode(
                            [
                                TextNode("Content type 2, value 1"),
                            ]
                        )
                    ),
                ],
                number=1,
                title=SentenceNode([TextNode("Some title 2.1")]),
                reference_anchor="ref-content_type2-1-XXYY",
                content_anchor="cnt-content_type2-1-XXYY",
            ),
        },
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == (
        "content_type1:value1:Content type 1, value 1:1:Some title 1.1:"
        "cnt-content_type1-1-XXYY:ref-content_type1-1-XXYY::content_type1:"
        "value2:Content type 1, value 2:2:Some title 1.2:"
        "cnt-content_type1-2-XXYY:ref-content_type1-2-XXYY::"
    )


def test_source_node():
    templates = {
        "text.j2": "{{ value }}",
        "callout.j2": "",
        "callouts_entry.j2": "{{ marker }} - {{ value }}",
        "source-default-somelang.j2": "The blocktype+language template",
        "source-somelang.j2": "The language template",
        "source-default.j2": "The blocktype template",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = SourceNode(
        language="somelang",
        code=[
            TextNode("import sys"),
            TextNode("import: os"),
            TextNode(""),
            TextNode('print(os.environ["HOME"])'),
        ],
        markers=[CalloutNode(1, "imp"), CalloutNode(3, "env")],
        callouts=[
            CalloutsEntryNode("imp", "This is an import"),
            CalloutsEntryNode("env", "Environment variables are paramount"),
        ],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == "The blocktype+language template"



# Inline nodes can be found inside a paragraph

from mau.nodes.nodes import Node, ValueNode


class WordNode(ValueNode):
    """This is a single word, it's used internally
    and eventually packed together with others into
    a TextNode
    """

    node_type = "word"


class TextNode(ValueNode):
    """This contains plain text and is created
    as a collation of multiple WordNode objects
    """

    node_type = "text"


class VerbatimNode(ValueNode):
    """This node contains verbatim text."""

    node_type = "verbatim"


class SentenceNode(Node):
    """A recursive container node.

    This node represents the content of a paragraph, but it is recursive,
    while ParagraphNode is not.
    """

    node_type = "sentence"

    def __init__(self, content):
        super().__init__()
        self.content = content

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "content": self.content,
        }


class StyleNode(Node):
    """Describes the style applied to a node."""

    node_type = "style"

    def __init__(self, value, content):
        super().__init__()
        self.value = value
        self.content = content

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "value": self.value,
            "content": self.content,
        }


class MacroNode(Node):
    """This node contains a macro, with a name and arguments."""

    node_type = "macro"

    def __init__(self, name, args=None, kwargs=None):
        super().__init__()
        self.name = name
        self.args = args or []
        self.kwargs = kwargs or {}

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "name": self.name,
            "args": self.args,
            "kwargs": self.kwargs,
        }


class ClassNode(Node):
    """Text with one or more classes."""

    node_type = "class"

    def __init__(self, classes, content):
        super().__init__()
        self.classes = classes
        self.content = content

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "classes": self.classes,
            "content": self.content,
        }


class LinkNode(Node):
    """This node contains a link."""

    node_type = "link"

    def __init__(self, target, text=None):
        super().__init__()
        self.target = target
        self.text = text

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "target": self.target,
            "text": self.text,
        }


class ImageNode(Node):
    """This node contains an inline image."""

    node_type = "image"

    def __init__(self, uri, alt_text=None, width=None, height=None):
        super().__init__()
        self.uri = uri
        self.alt_text = alt_text
        self.width = width
        self.height = height

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "uri": self.uri,
            "alt_text": self.alt_text,
            "width": self.width,
            "height": self.height,
        }


class ListItemNode(Node):
    """An entry in a list."""

    node_type = "list_item"

    def __init__(self, level, content):
        super().__init__()
        self.level = level
        self.content = content

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "level": self.level,
            "content": self.content,
        }



from typing import Optional

from mau.text_buffer.context import Context

Position = tuple[int, int]

# The TextBuffer is an object used to interact with a text file.
#
# The object contains by default an empty piece of text and no
# initial context, but it can accept both at initialisation time.
#
# The object itself doesnt interact with files, the text content
# has to be loaded externally. The `text` parameter is a single
# string containing newlines, the object will split it internally.
# When a piece of text is loaded it is split into lines using `\n`.
#
# The attribute `initial_context` allows to nest pieces of text that
# come from different sources. The property `context` takes the
# initial context into account and adjusts the line and column.
#
# The object has several properties:
# * `eof` - True if the current line is beyond the last line.
# * `eol` - True if the current column is beyond the last column
#   of the current line.
# * `current_line` - The current line of the file or an empty string
#   in case of EOF.
# * `current_char` - The current character of the current line
#   or an empty string in case of EOL/EOF.
# * `peek_char` - The next character of the current line
#   or an empty string in case of EOL/EOF.
# * `tail` - The remaining part of the line starting with the next
#   character.
# * `position` - (getter/setter) A tuple (line, column) with the
#   current position.
# * `context` - A instance of Context relative to `initial_context`.
#
# The class also exposes three main methods:
# * `load` - Loads a string into the internal buffer. The string is
#   split using newlines `\n`.
# * `nextline` - Moves to the beginning of the next line.
# * `skip` - Skips the given number of characters.
#
# The class has been designed to minimise the number of exceptions it
# can raise, so it will quietly go beyond the end of the line of beyond
# the end of the file. It's the users task to check whether one of
# these conditions occured.


class TextBuffer:
    def __init__(self, text: str = "", context: Optional[Context] = None):
        self.lines: list[str] = []
        self.load(text, context)

    def load(self, text: str, context: Optional[Context] = None) -> None:
        self.lines = text.split("\n") if text != "" else []

        self.initial_context = context or Context()
        self.line = 0
        self.column = 0

    @property
    def context(self) -> Context:
        return Context(
            line=self.line + self.initial_context.line,
            column=self.column + self.initial_context.column,
            source=self.initial_context.source,
            text=self.current_line,
        )

    @property
    def eof(self) -> bool:
        """
        True if the position is beyond EOF.
        """
        return self.line >= len(self.lines)

    @property
    def eol(self):
        """
        True if the position is beyond EOL.
        """
        return self.column >= len(self.current_line)

    @property
    def current_line(self):
        """
        Returns the current line.

        This property returns the current line without advancing the index.
        If the buffer is reading after the last line it returns an empty string.
        """
        try:
            return self.lines[self.line]
        except IndexError:
            return ""

    @property
    def current_char(self):
        """
        Returns the current character.

        This property returns the current character without advancing the index.
        If the buffer is reading after the last character of the line
        or after the last line it returns an empty string.
        """
        try:
            return self.current_line[self.column]
        except IndexError:
            return ""

    @property
    def peek_char(self):
        """
        Returns the next character.

        This property returns the next character without advancing the index.
        If the buffer is reading after the last character of the line
        or after the last line it returns an empty string.
        """

        try:
            return self.current_line[self.column + 1]
        except IndexError:
            return ""

    @property
    def tail(self):
        """
        Returns the remaining part of the line.

        This property returns a string with the last part of the current
        line from the current character to the end.
        """
        return self.current_line[self.column :]

    @property
    def position(self):
        """
        Returns a tuple with the current position.
        """
        return (self.line, self.column)

    @position.setter
    def position(self, position: Position):
        """
        Sets the current position.
        """
        self.line, self.column = position

    def nextline(self):
        """
        Moves the index to the beginning of the next line
        """
        if self.line <= len(self.lines):
            self.line += 1

        self.column = 0

    def skip(self, chars=1):
        """
        Skips the given number of characters (default 1). Can silently
        go over the end of the line.
        """
        new_column = self.column + chars

        if new_column > len(self.current_line):
            new_column = len(self.current_line)

        self.column = new_column



import itertools

from mau.lexers.base_lexer import Token, TokenTypes
from mau.lexers.text_lexer import TextLexer
from mau.nodes.footnotes import FootnoteNode
from mau.nodes.inline import (
    ClassNode,
    ImageNode,
    LinkNode,
    MacroNode,
    SentenceNode,
    StyleNode,
    TextNode,
    VerbatimNode,
    WordNode,
)
from mau.nodes.references import ReferenceNode
from mau.parsers.arguments import set_names_and_defaults
from mau.parsers.arguments_parser import ArgumentsParser
from mau.parsers.base_parser import BaseParser

# This is a simple map to keep track of the official
# name of styles introduced by special characters
MAP_STYLES = {"_": "underscore", "*": "star", "^": "caret", "~": "tilde"}


# The TextParser is a nested parser.
# The parsing always starts with parse_sentence
# and from there all components of the text are explored
class TextParser(BaseParser):
    lexer_class = TextLexer

    def __init__(self, tokens):
        super().__init__(tokens)

        # These are the footnotes found in this text
        self.footnotes = {}

        # These are the references found in this text
        # The format of this dictionary is
        # {("content_type", "name"): node}
        self.references = {}

    def _process_functions(self):
        return [self._process_eol, self._process_sentence]

    def _process_eol(self):
        # This simply parses the end of line.

        self._get_token(TokenTypes.EOL)

        return True

    def _process_sentence(self):
        self._save(self._parse_sentence())

        return True

    def _parse_word(self):
        """
        Parse a single word.
        """
        return WordNode(self._get_token().value)

    def _parse_style(self):
        """
        Parse a sentence surrounded by style markers.
        """

        # Get the style marker
        style = self._get_token(
            TokenTypes.LITERAL, check=lambda x: x in MAP_STYLES
        ).value

        # Get everything until the next marker
        content = self._parse_sentence(stop_tokens={Token(TokenTypes.LITERAL, style)})

        # Get the closing marker
        self._get_token(TokenTypes.LITERAL, style)

        return StyleNode(MAP_STYLES[style], content)

    def _parse_escape(self):
        """
        Parse an escaped element.
        """

        # Drop the backslash
        self._get_token(TokenTypes.LITERAL, "\\")

        return WordNode(self._get_token().value)

    def _parse_styled_text(self, stop_tokens=None):
        """
        Parse multiple possible elements: escapes, classes,
        macros, verbatim, styles, links, words.
        This is a helper for the function _parse_sentence
        that takes into account all possible elements of
        syntax, stopping if the token is among the listed ones.
        """
        stop_tokens = stop_tokens or set()

        if self._peek_token() in stop_tokens:
            return None

        with self:
            return self._parse_escape()

        with self:
            return self._parse_macro()

        with self:
            return self._parse_verbatim()

        with self:
            return self._parse_style()

        return self._parse_word()

    def _parse_sentence(self, stop_tokens=None):
        """
        Parse a sentence, which is made of multiple
        elements identified by _parse_styled_text, until
        the EOF, the EOL, or a specific set of tokens
        passed as argument.
        """

        content = []
        stop_tokens = stop_tokens or set()
        stop_tokens = stop_tokens.union({Token(TokenTypes.EOF), Token(TokenTypes.EOL)})

        result = self._parse_styled_text(stop_tokens)
        while result is not None:
            content.append(result)
            result = self._parse_styled_text(stop_tokens)

        # Group consecutive WordNode nodes into a single TextNode
        grouped_nodes = []
        for key, group in itertools.groupby(content, lambda x: x.__class__ == WordNode):
            if key:
                text = "".join([n.value for n in group])
                grouped_nodes.append(TextNode(text))
            else:
                grouped_nodes.extend(list(group))

        return SentenceNode(content=grouped_nodes)

    def _parse_verbatim(self):
        """
        Parse text in `verbatim`.
        """

        # Get the verbatim marker
        self._get_token(TokenTypes.LITERAL, "`")

        # Get the content tokens until the
        # next verbatim marker or EOL
        content = self._collect_join(
            [Token(TokenTypes.LITERAL, "`"), Token(TokenTypes.EOL)],
        )

        # Remove the closing marker
        self._get_token(TokenTypes.LITERAL, "`")

        return VerbatimNode(content)

    def _parse_macro_link(self, args, kwargs):
        """
        Parse a link macro in the form [link](target, text).
        """

        args, kwargs = set_names_and_defaults(
            args, kwargs, ["target", "text"], {"text": None}
        )

        target = kwargs.get("target")

        text = kwargs["text"]
        if text is None:
            text = target

        return LinkNode(target=target, text=text)

    def _parse_macro_mailto(self, args, kwargs):
        """
        Parse a mailto macro in the form [mailto](email).
        """

        args, kwargs = set_names_and_defaults(
            args, kwargs, ["email", "text"], {"text": None}
        )

        email = kwargs.get("email")
        text = kwargs.get("text")

        target = f"mailto:{email}"
        if text is None:
            text = email

        return LinkNode(target, text)

    def _parse_macro_class(self, args, kwargs):
        """
        Parse a class macro in the form [class](text, "class1,class2,...").
        """

        args, kwargs = set_names_and_defaults(
            args, kwargs, ["text", "classes"], {"classes": ""}
        )

        current_context = self._current_token.context

        text = kwargs.get("text")

        par = self.analyse(text, current_context)

        # Text should return a single sentence node
        result = par.nodes[0]

        # Multiple classes are separated by commas
        classes = kwargs["classes"].split(",")

        return ClassNode(classes, result)

    def _parse_macro_image(self, args, kwargs):
        """
        Parse an inline image macro in the form
        [image](uri, alt_text, width, height).
        """

        # Assign names and get the arguments
        args, kwargs = set_names_and_defaults(
            args,
            kwargs,
            ["uri", "alt_text", "width", "height"],
            {"alt_text": None, "width": None, "height": None},
        )

        return ImageNode(
            uri=kwargs.get("uri"),
            alt_text=kwargs.get("alt_text"),
            width=kwargs.get("width"),
            height=kwargs.get("height"),
        )

    def _parse_macro_footnote(self, args, kwargs):
        """
        Parse a footnote macro in the form
        [footnote](name).
        """

        args, kwargs = set_names_and_defaults(args, kwargs, ["name"])

        name = kwargs["name"]

        node = FootnoteNode()

        self.footnotes[name] = node

        return node

    def _parse_macro_reference(self, args, kwargs):
        """
        Parse a reference macro in the form
        [reference](type, name, category).
        """

        args, kwargs = set_names_and_defaults(
            args,
            kwargs,
            ["type", "name", "category"],
            {"category": None},
        )

        content_type = kwargs["type"]
        name = kwargs["name"]
        category = kwargs["category"]

        node = ReferenceNode(content_type, name, category=category)

        self.references[(content_type, name)] = node

        return node

    def _collect_macro_args(self):
        self._get_token(TokenTypes.LITERAL, "(")

        all_args = []

        while not (
            self._peek_token_is(TokenTypes.LITERAL, ")")
            or self._peek_token_is(TokenTypes.EOL)
        ):
            if self._peek_token_is(TokenTypes.LITERAL, '"'):
                self._get_token(TokenTypes.LITERAL, '"')

                value = self._collect_join(
                    stop_tokens=[
                        Token(TokenTypes.LITERAL, '"'),
                        Token(TokenTypes.EOL),
                    ],
                )

                self._get_token(TokenTypes.LITERAL, '"')

                arguments = f'"{value}"'
            else:
                arguments = self._collect_join(
                    stop_tokens=[
                        Token(TokenTypes.LITERAL, ","),
                        Token(TokenTypes.LITERAL, ")"),
                        Token(TokenTypes.EOL),
                    ],
                )

            all_args.append(arguments)

        self._get_token(TokenTypes.LITERAL, ")")

        return "".join(all_args)

    def _parse_macro(self):
        self._get_token(TokenTypes.LITERAL, "[")
        macro_name = self._get_token(TokenTypes.TEXT).value
        self._get_token(TokenTypes.LITERAL, "]")

        arguments = self._collect_macro_args()

        current_context = self._peek_token().context

        par = ArgumentsParser.analyse(arguments, current_context)

        args, kwargs, _ = par.process_arguments()

        if macro_name == "link":
            return self._parse_macro_link(args, kwargs)

        if macro_name == "mailto":
            return self._parse_macro_mailto(args, kwargs)

        if macro_name == "image":
            return self._parse_macro_image(args, kwargs)

        if macro_name == "footnote":
            return self._parse_macro_footnote(args, kwargs)

        if macro_name == "reference":
            return self._parse_macro_reference(args, kwargs)

        if macro_name == "class":
            return self._parse_macro_class(args, kwargs)

        return MacroNode(macro_name, args=args, kwargs=kwargs)



# pylint: disable=too-many-lines

import hashlib
import re

from mau.lexers.base_lexer import TokenTypes as BLTokenTypes
from mau.lexers.main_lexer import MainLexer
from mau.lexers.main_lexer import TokenTypes as MLTokenTypes
from mau.nodes.footnotes import CommandFootnotesNode, FootnotesEntryNode
from mau.nodes.inline import ListItemNode, TextNode
from mau.nodes.page import (
    BlockNode,
    CommandTocNode,
    ContentImageNode,
    ContentNode,
    HeaderNode,
    HorizontalRuleNode,
    ListNode,
    ParagraphNode,
)
from mau.nodes.references import CommandReferencesNode, ReferencesEntryNode
from mau.nodes.source import CalloutNode, CalloutsEntryNode, SourceNode
from mau.parsers.arguments_parser import ArgumentsParser
from mau.parsers.base_parser import BaseParser
from mau.parsers.environment import Environment
from mau.parsers.preprocess_variables_parser import PreprocessVariablesParser
from mau.parsers.text_parser import TextParser
from mau.tokens.tokens import Token


def header_anchor(text, level):
    """
    Return a sanitised anchor for a header.
    """

    # Everything lowercase
    sanitised_text = text.lower()

    # Get only letters, numbers, dashes, spaces, and dots
    sanitised_text = "".join(re.findall("[a-z0-9-\\. ]+", sanitised_text))

    # Remove multiple spaces
    sanitised_text = "-".join(sanitised_text.split())

    hashed_value = hashlib.md5(f"{level} {text}".encode("utf-8")).hexdigest()[:4]

    return f"{sanitised_text}-{hashed_value}"


def footnote_anchor(content):
    return hashlib.md5(str(content).encode("utf-8")).hexdigest()[:8]


def reference_anchor(content):
    return hashlib.md5(str(content).encode("utf-8")).hexdigest()[:8]


# The MainParser is in charge of parsing
# the whole input, calling other parsers
# to manage single paragraphs or other
# things like variables.
class MainParser(BaseParser):
    lexer_class = MainLexer

    def __init__(
        self, tokens, environment=None, references=None, reference_entries=None
    ):
        super().__init__(tokens)

        self.environment = environment or Environment()

        self.headers = []

        # This dictionary containes the footnotes created
        # in the text through a macro.
        self.footnotes = {}

        # This dictionary contains the content of each
        # footnote created through blocks.
        # This is a helper dictionary that will be merged
        # with self.footnotes once the parsing is completed.
        self.footnotes_data = {}

        # This list contains all the footnote entries
        # that will be shown by a footnotes command.
        self.footnote_entries = []

        # This dictionary containes the references created
        # in the text through a macro.
        self.reference_mentions = {}

        # This dictionary contains the content of each
        # reference created through blocks.
        self.reference_data = {}

        # This list contains all the references contained
        # in this parser in the form
        # {content_type:[references]}.
        self.references = references if references is not None else {}

        # This list contains all the reference entries
        # that will be shown by a references command in the form
        # {content_type:[references]}.
        # This is a copy of self.references that however
        # contains ReferencesEntryNode nodes.
        self.reference_entries = (
            reference_entries if reference_entries is not None else {}
        )

        # When we define a block we establish an alias
        # {alias:actual_block_name}.
        self.block_aliases = {}

        # Each block we define can have default values
        # {actual_block_name:kwargs}.
        self.block_defaults = {}

        # Each block we define can have names for unnamed arguments
        # {actual_block_name:kwargs}.
        self.block_names = {}

        # This establishes a default block definition so that
        # [source] = [source, engine=source]
        # This definition can be overridden by custom block definitions.
        self.block_aliases["source"] = "default"
        self.block_defaults["source"] = {"engine": "source", "language": "text"}
        self.block_names["source"] = ["language"]

        self.block_aliases["footnote"] = "default"
        self.block_defaults["footnote"] = {"engine": "footnote"}
        self.block_names["footnote"] = ["name"]

        self.block_aliases["reference"] = "default"
        self.block_defaults["reference"] = {"engine": "reference"}
        self.block_names["reference"] = ["type", "name"]

        self.block_aliases["admonition"] = "admonition"
        self.block_names["admonition"] = ["class", "icon", "label"]

        # Iterate through block definitions passed as variables
        # for alias, block_definition in (
        #     self.variables["mau"].get("block_definitions", {}).items()
        # ):
        #     try:
        #         blocktype = block_definition["blocktype"]
        #         self.block_aliases[alias] = blocktype
        #     except KeyError:
        #         raise ConfigurationError(
        #             f"Block definition '{alias}' is missing key 'blocktype'"
        #         )

        #     try:
        #         self.block_defaults[blocktype] = block_definition["kwargs"]
        #     except KeyError:
        #         raise ConfigurationError(
        #             f"Block definition '{alias}' is missing key 'kwargs'"
        #         )

        # This is a buffer for a block title
        self._title = None

        # This is the function used to create the header
        # anchors. It can be specified through
        # mau.header_anchor_function to override
        # the default one.
        self.header_anchor = self.environment.getvar(
            "mau.header_anchor_function", header_anchor
        )

        # A temporary space to store parsed arguments
        # The tuple represents (args, kwargs, tags)
        self.arguments = ([], {}, [])

    def create_footnotes(self):
        for num, footnote in enumerate(self.footnotes.values(), start=1):
            footnote.number = num

        for key, footnote in self.footnotes.items():
            data = self.footnotes_data[key]
            footnote.content = data["content"]
            anchor = footnote_anchor(footnote.content)

            footnote.reference_anchor = f"fr-{anchor}-{footnote.number}"
            footnote.content_anchor = f"fd-{anchor}-{footnote.number}"

            self.footnote_entries.append(
                FootnotesEntryNode(
                    content=footnote.content,
                    number=footnote.number,
                    reference_anchor=footnote.reference_anchor,
                    content_anchor=footnote.content_anchor,
                )
            )

    def process_references(self):
        # Example of stored content
        # self.references = {
        #  (type1, name1) = node1
        #  (type1, name2) = node2
        #  (type2, name3) = node3
        # }
        #
        # self.reference_data = {
        #  (type1, name1) = content
        #  (type1, name2) = content
        #  (type2, name3) = content
        # }

        content_types = {i[0] for i in self.reference_data}
        content_types_num = {}

        for content_type in content_types:
            content_types_num[content_type] = 1

        for key, value in self.reference_mentions.items():
            data = self.reference_data[key]
            content_type = value.content_type

            value.content = data["content"]
            value.title = data["title"]
            anchor = reference_anchor(value.content)

            value.number = content_types_num[content_type]
            content_types_num[content_type] += 1

            value.reference_anchor = f"ref-{content_type}-{value.number}-{anchor}"
            value.content_anchor = f"cnt-{content_type}-{value.number}-{anchor}"

            self.references[key] = value
            self.reference_entries[key] = ReferencesEntryNode(
                content_type=value.content_type,
                name=value.name,
                category=value.category,
                content=value.content,
                number=value.number,
                title=value.title,
                reference_anchor=value.reference_anchor,
                content_anchor=value.content_anchor,
            )

    def _pop_arguments(self):
        # This return the arguments and resets the
        # cached ones.
        args, kwargs, tags = self.arguments
        self.arguments = ([], {}, [])

        return args, kwargs, tags

    def _push_arguments(self, args, kwargs, tags):
        self.arguments = (args, kwargs, tags)

    def _process_functions(self):
        # All the functions that this parser provides.

        return [
            self._process_eol,
            self._process_horizontal_rule,
            self._process_single_line_comment,
            self._process_multi_line_comment,
            self._parse_variable_definition,
            self._process_command,
            self._process_title,
            self._process_arguments,
            self._process_header,
            self._process_block,
            self._process_include_content,
            self._process_list,
            self._process_paragraph,
        ]

    def _pop_title(self):
        # This return the title and resets the
        # cached one, so no other block will
        # use it.
        title = self._title
        self._title = None
        return title

    def _push_title(self, title):
        # When we parse a title we can store it here
        # so that it is available to the next block
        # that will use it.
        self._title = title

    def _collect_text_content(self):
        # Collects all adjacent text tokens
        # into a single string

        if not self._peek_token_is(BLTokenTypes.TEXT):  # pragma: no cover
            return None

        values = []

        # Get all tokens
        while self._peek_token_is(BLTokenTypes.TEXT):
            values.append(self._get_token().value)
            self._get_token(BLTokenTypes.EOL)

        return " ".join(values)

    def _parse_text_content(self, text):
        # Parse a piece of text using the TextParser.

        current_context = self._current_token.context

        # Replace variables
        preprocess_parser = PreprocessVariablesParser.analyse(
            text,
            current_context,
            self.environment,
        )
        text = preprocess_parser.nodes[0].value

        # Parse the text
        text_parser = TextParser.analyse(
            text,
            current_context,
        )

        # A text parser returns a single sentence node
        result = text_parser.nodes[0]

        # Store the footnotes
        self.footnotes.update(text_parser.footnotes)

        # The format of the stored content dictionary is
        # {(content_type,name): node}
        self.reference_mentions.update(text_parser.references)

        return result

    def _process_eol(self):
        # This simply parses the end of line.

        self._get_token(BLTokenTypes.EOL)

        return True

    def _process_horizontal_rule(self):
        # The horizontal rule ---

        self._get_token(MLTokenTypes.HORIZONTAL_RULE)

        args, kwargs, tags = self._pop_arguments()

        self._save(HorizontalRuleNode(args, kwargs, tags))

        return True

    def _process_single_line_comment(self):
        # // A comment on a single line

        self._get_token(MLTokenTypes.COMMENT)
        self._get_token(BLTokenTypes.EOL)

        return True

    def _process_multi_line_comment(self):
        # ////
        # A comment
        # on multiple lines
        # ////

        self._get_token(MLTokenTypes.MULTILINE_COMMENT)
        self._collect([Token(MLTokenTypes.MULTILINE_COMMENT, "////")])
        self._force_token(MLTokenTypes.MULTILINE_COMMENT)

        return True

    def _parse_variable_definition(self):
        # This parses the definition of a variable
        #
        # Simple variables are defined as :name:value
        # as True booleans as just :name:
        # and as False booleas as :!name:
        #
        # Variable names can use a namespace with
        # :namespace.name:value

        # Get the mandatory variable name
        self._get_token(MLTokenTypes.VARIABLE, ":")
        variable_name = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.LITERAL, ":")

        # Assume the variable is a flag
        variable_value = True

        # If the name starts with ! it's a false flag
        if variable_name.startswith("!"):
            variable_value = False
            variable_name = variable_name[1:]

        # Get the optional value
        value = self._collect_join([Token(BLTokenTypes.EOL)])

        # The value is assigned only if the variable
        # is not a negative flag. In that case it is ignored
        if variable_value and len(value) > 0:
            variable_value = value

        self.environment.setvar(variable_name, variable_value)

        return True

    def _process_command(self):
        # Parse a command in the form ::command:arguments

        self._get_token(MLTokenTypes.COMMAND, "::")
        name = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.LITERAL, ":")

        args = []
        kwargs = {}
        tags = []

        # Commands can have arguments
        arguments = ""
        with self:
            arguments = self._get_token(BLTokenTypes.TEXT).value

            current_context = self._current_token.context

            arguments_parser = ArgumentsParser.analyse(
                arguments,
                current_context,
            )

            args, kwargs, tags = arguments_parser.process_arguments()

        if name == "defblock":
            # Block definitions must have at least 2 arguments,
            # the alias and the block type.
            if len(args) < 2:
                self._error(
                    "Block definitions require at least two unnamed arguments: ALIAS and BLOCKTYPE"
                )

            block_alias = args.pop(0)
            block_type = args.pop(0)

            self.block_aliases[block_alias] = block_type
            self.block_defaults[block_alias] = kwargs
            self.block_names[block_alias] = args

        elif name == "toc":
            self._save(
                CommandTocNode(
                    entries=self.headers, args=args, kwargs=kwargs, tags=tags
                )
            )

        elif name == "footnotes":
            self._save(
                CommandFootnotesNode(
                    entries=self.footnote_entries, args=args, kwargs=kwargs, tags=tags
                )
            )

        elif name == "references":
            # Assign names
            args, kwargs = self._set_names_and_defaults(
                args,
                kwargs,
                ["content_type", "category", "name"],
                {"category": None, "name": None},
            )

            content_type = kwargs.pop("content_type")
            category = kwargs.pop("category")
            name = kwargs.pop("name")

            self._save(
                CommandReferencesNode(
                    entries=self.reference_entries,
                    content_type=content_type,
                    name=name,
                    category=category,
                    args=args,
                    kwargs=kwargs,
                    tags=tags,
                )
            )

        return True

    def _process_title(self):
        # Parse a title in the form
        #
        # . This is a title
        # or
        # .This is a title

        # Parse the mandatory dot
        self._get_token(MLTokenTypes.TITLE, ".")

        # Parse the optional white spaces
        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Get the text of the title
        text = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.EOL)

        current_context = self._current_token.context

        # Titles can contain Mau code
        text_parser = TextParser.analyse(text, current_context)

        title = text_parser.nodes[0]

        self._push_title(title)

        return True

    def _process_arguments(self):
        # Parse arguments in the form
        # [unnamed1, unnamed2, ..., named1=value1, name2=value2, ...]

        self._get_token(MLTokenTypes.ARGUMENTS, "[")
        text = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.LITERAL, "]")

        current_context = self._current_token.context

        preprocess_parser = PreprocessVariablesParser.analyse(
            text,
            current_context,
            environment=self.environment,
        )
        text = preprocess_parser.nodes[0].value

        # Parse the text
        arguments_parser = ArgumentsParser.analyse(
            text,
            current_context,
        )

        args, kwargs, tags = arguments_parser.process_arguments()
        self._push_arguments(args, kwargs, tags)

        return True

    def _process_header(self):
        # Parse a header in the form
        #
        # = Header
        #
        # The number of equal signs is arbitrary
        # and represents the level of the header.
        # Headers are automatically assigned an anchor
        # created using the provided function self.header_anchor

        # Get all the equal signs
        header = self._get_token(MLTokenTypes.HEADER).value

        # Get the mandatory white spaces
        self._get_token(BLTokenTypes.WHITESPACE)

        # Get the text of the header and calculate the level
        text = self._get_token(BLTokenTypes.TEXT).value
        level = len(header)

        # Generate the anchor and append it to the TOC
        anchor = self.header_anchor(text, level)

        # Consume the arguments
        args, kwargs, tags = self._pop_arguments()

        # Generate the header node
        header_node = HeaderNode(
            value=text,
            level=str(level),
            anchor=anchor,
            args=args,
            tags=tags,
            kwargs=kwargs,
        )

        self.headers.append(header_node)

        self._save(header_node)

        return True

    def _collect_lines(self, stop_tokens):
        # This collects several lines of text in a list
        # until it gets to a line that begins with one
        # of the tokens listed in stop_tokens.
        # It is useful for block or other elements that
        # are clearly surrounded by delimiters.
        lines = []

        while self._peek_token() not in stop_tokens:
            lines.append(self._collect_join([Token(BLTokenTypes.EOL)]))
            self._get_token(BLTokenTypes.EOL)

        return lines

    def _process_block(
        self,
    ):  # pylint: disable=too-many-statements, too-many-branches, too-many-locals
        # Parse a block in the form
        #
        # [block_type]
        # ----
        # Content
        # ----
        # Optional secondary content
        #
        # Blocks are delimited by 4 consecutive identical characters.

        # Get the delimiter and check the length
        delimiter = self._get_token(MLTokenTypes.BLOCK).value

        self._get_token(BLTokenTypes.EOL)

        # Collect everything until the next delimiter
        content = self._collect_lines(
            [Token(MLTokenTypes.BLOCK, delimiter), Token(BLTokenTypes.EOF)]
        )

        self._force_token(MLTokenTypes.BLOCK, delimiter)
        self._get_token(BLTokenTypes.EOL)

        # Get the optional secondary content
        secondary_content = self._collect_lines(
            [Token(BLTokenTypes.EOL), Token(BLTokenTypes.EOF)]
        )

        if delimiter in secondary_content:
            # This probably means that the input contains an error
            # and we are in a situation like

            # ----
            #
            # ----
            # Text
            # ----

            # Where ["Text", "----"] is considered the secondary content
            # of an empty block.
            self._error(
                "Detected unclosed block (possibly before this line)"
            )  # pragma: no cover

        # Consume the title
        title = self._pop_title()

        # Consume the arguments
        args, kwargs, _ = self._pop_arguments()

        # The first unnamed argument is the block type
        try:
            blocktype = args.pop(0)
        except IndexError:
            blocktype = "default"

        # If there is a block alias for blocktype replace it
        # otherwise use the blocktype we already have

        # Retrieve the block names and defaults for the
        # specific type of block

        block_names = self.block_names.get(blocktype, [])
        block_defaults = self.block_defaults.get(blocktype, {})

        # Now replace the alias with the true block type
        blocktype = self.block_aliases.get(blocktype, blocktype)

        # Assign names
        args, kwargs = self._set_names_and_defaults(
            args,
            kwargs,
            block_names,
            block_defaults,
        )

        # Extract classes and convert them into a list
        classes = []
        if "classes" in kwargs:
            classes = kwargs.pop("classes")

            if classes:
                classes = classes.split(",")

        # Extract condition if present and process it
        condition = kwargs.pop("condition", None)

        # Run this only if there is a condition on this block
        if condition is not None:
            try:
                # The condition should be either test:variable:value or test:variable:
                test, variable, value = condition.split(":")
            except ValueError:
                self._error(
                    (
                        f"Condition {condition} is not in the form"
                        '"test:variable:value" or "test:variable:'
                    )
                )

            # If there is no value use True
            if len(value) == 0:
                value = True

            # Check if the variable matches the value and apply the requested test
            match = self.environment.getvar(variable) == value
            result = test == "if"

            # If the condition is not satisfied return
            if match is not result:
                return True

        # Extract the preprocessor
        preprocessor = kwargs.pop("preprocessor", "none")

        # Extract the engine
        engine = kwargs.pop("engine", "default")

        if engine == "source":
            # Engine "source" extracts the content (source code),
            # the callouts, and the highlights.
            # The default language is "text".

            self._parse_source_engine(
                blocktype, content, secondary_content, title, kwargs
            )

            return True

        if engine == "footnote":
            name = kwargs.pop("name")

            current_context = self._current_token.context

            environment = self.environment

            content_parser = MainParser.analyse(
                "\n".join(content),
                current_context,
                environment=environment,
                references=self.references,
                reference_entries=self.reference_entries,
            )

            content = content_parser.nodes

            self.footnotes_data[name] = {
                "content": content,
            }

            return True

        if engine == "reference":
            content_type = kwargs["type"]
            name = kwargs["name"]

            current_context = self._current_token.context

            environment = self.environment

            content_parser = MainParser.analyse(
                "\n".join(content),
                current_context,
                environment=environment,
                references=self.references,
                reference_entries=self.reference_entries,
            )

            content = content_parser.nodes

            self.reference_data[(content_type, name)] = {
                "content": content,
                "title": title,
            }

            return True

        # Create the node parameters according to the engine
        if engine == "raw":
            # Engine "raw" doesn't process the content,
            # so we just pass it untouched in the form of
            # a TextNode per line.
            content = [TextNode(line) for line in content]
            secondary_content = [TextNode(line) for line in secondary_content]
        elif engine in ["default", "mau"]:
            # Both engines parse content and secondary content
            # using a new parser but the default one should merge
            # headers and footnotes into the current one.

            current_context = self._current_token.context

            if engine == "default":
                environment = self.environment
            else:
                environment = Environment()

            content_parser = MainParser.analyse(
                "\n".join(content),
                current_context,
                environment=environment,
                references=self.references,
                reference_entries=self.reference_entries,
            )

            secondary_content_parser = MainParser.analyse(
                "\n".join(secondary_content),
                current_context,
                environment=environment,
            )

            content = content_parser.nodes
            secondary_content = secondary_content_parser.nodes

            if engine == "default":
                self.footnotes.update(content_parser.footnotes)
                self.footnotes_data.update(content_parser.footnotes_data)
                self.headers.extend(content_parser.headers)
                self.reference_mentions.update(content_parser.reference_mentions)
                self.reference_data.update(content_parser.reference_data)
            else:
                content_parser.create_footnotes()
                content_parser.process_references()
        else:
            self._error(f"Engine {engine} is not available")

        self._save(
            BlockNode(
                blocktype=blocktype,
                content=content,
                secondary_content=secondary_content,
                classes=classes,
                title=title,
                engine=engine,
                preprocessor=preprocessor,
                args=args,
                kwargs=kwargs,
            )
        )

        return True

    def _parse_source_engine(
        self, blocktype, content, secondary_content, title, kwargs
    ):  # pylint: disable=too-many-locals
        # Parse a source block in the form
        #
        # [source, language, attributes...]
        # ----
        # content
        # ----
        #
        # Source blocks support the following attributes
        #
        # callouts=":" The separator used by callouts
        # highlight="@" The special character to turn on highlight
        #
        # [source, language, attributes...]
        # ----
        # content:1:
        # ----
        #
        # [source, language, attributes...]
        # ----
        # content:@:
        # ----
        #
        # Callout descriptions can be added to the block
        # as secondary content with the syntax
        #
        # [source, language, attributes...]
        # ----
        # content:name:
        # ----
        # <name>: <description>
        #
        # Since Mau uses Pygments, the attribute language
        # is one of the langauges supported by that tool.

        # Get the delimiter for callouts (":" by default)
        delimiter = kwargs.pop("callouts", ":")

        # A list that contains callout markers in
        # the form (linenum,name)
        callout_markers = []

        # Get the marker for highlighted lines ("@" by default)
        highlight_marker = kwargs.pop("highlight", "@")

        # A list of highlighted lines
        highlighted_lines = []

        # Source blocks preserve anything is inside

        # This is a list of all lines that might contain
        # a callout. They will be further processed
        # later to be sure.
        lines_with_callouts = [
            (linenum, line)
            for linenum, line in enumerate(content)
            if line.endswith(delimiter)
        ]

        # Each line in the previous list is processed
        # and stored if it contains a callout
        for linenum, line in lines_with_callouts:
            # Remove the final delimiter
            line = line[:-1]

            splits = line.split(delimiter)
            if len(splits) < 2:
                # It's a trap! There are no separators left
                continue

            # Get the callout and the line
            callout_name = splits[-1]
            line = delimiter.join(splits[:-1])

            content[linenum] = line

            # Check if we want to just highlight the line
            if callout_name == highlight_marker:
                highlighted_lines.append(linenum)
            else:
                callout_markers.append(CalloutNode(linenum, callout_name))

        # A list of CalloutEntryNode objects that contain the
        # text for each marker
        callout_contents = []

        # If there was secondary content it should be formatted
        # with callout names followed by colon and the
        # callout text.
        for line in secondary_content:
            if ":" not in line:
                self._error(
                    (
                        "Callout description should be written"
                        f"as 'name: text'. Missing ':' in '{line}'"
                    )
                )

            name, text = line.split(":")

            text = text.strip()

            callout_contents.append(CalloutsEntryNode(name, text))

        # Source blocks must preserve the content literally.
        # However, we need to remove escape characters from directives.
        # Directives are processed by the lexer, so if we want to
        # prevent Mau from interpreting them we have to escape them.
        # Escape characters are preserved by source blocks as anything
        # else, but in this case the character should be removed.
        textlines = []
        for line in content:
            if line.startswith(r"\::#"):
                line = line[1:]

            textlines.append(TextNode(line))

        self._save(
            SourceNode(
                blocktype=blocktype,
                code=textlines,
                language=kwargs["language"],
                callouts=callout_contents,
                highlights=highlighted_lines,
                markers=callout_markers,
                title=title,
            )
        )

    def _process_include_content(self):
        # Parse include content in the form
        #
        # << content_type:arguments

        # Get the mandatory "<<"
        self._get_token(MLTokenTypes.INCLUDE)

        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Get the content type
        content_type = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.LITERAL, ":")

        title = self._pop_title()

        args = []
        kwargs = {}
        tags = []

        # Commands can have arguments
        arguments = ""
        with self:
            arguments = self._get_token(BLTokenTypes.TEXT).value

            current_context = self._current_token.context

            arguments_parser = ArgumentsParser.analyse(
                arguments,
                current_context,
            )

            args, kwargs, tags = arguments_parser.process_arguments()

        if content_type == "image":
            return self._parse_content_image(title, args, kwargs, tags)

        return self._parse_standard_content(content_type, title, args, kwargs, tags)

    def _parse_content_image(self, title, args, kwargs, tags):
        # Parse a content image in the form
        #
        # << image:uri,alt_text,classes
        #
        # alt_text is the alternate text to use is the image is not reachable
        # and classes is a comma-separated list of classes

        # Consume the arguments
        args, kwargs = self._set_names_and_defaults(
            args,
            kwargs,
            ["uri", "alt_text", "classes"],
            {"alt_text": None, "classes": None},
        )

        uri = kwargs.pop("uri")
        alt_text = kwargs.pop("alt_text")
        classes = kwargs.pop("classes")

        if classes:
            classes = classes.split(",")

        self._save(
            ContentImageNode(
                uri=uri,
                alt_text=alt_text,
                classes=classes,
                title=title,
                kwargs=kwargs,
                tags=tags,
            )
        )

        return True

    def _parse_standard_content(self, content_type, title, args, kwargs, tags):
        # This is the fallback for an unknown content type

        self._save(
            ContentNode(
                content_type=content_type,
                title=title,
                args=args,
                kwargs=kwargs,
                tags=tags,
            )
        )

        return True

    def _process_list(self):
        # Parse a list.
        # Lists can be ordered (using numbers)
        #
        # * One item
        # * Another item
        #
        # or unordered (using bullets)
        #
        # # Item 1
        # # Item 2
        #
        # The number of headers increases
        # the depth of each item
        #
        # # Item 1
        # ## Sub-Item 1.1
        #
        # Spaces before and after the header are ignored.
        # So the previous list can be also written
        #
        # # Item 1
        #   ## Sub-Item 1.1
        #
        # Ordered and unordered lists can be mixed.
        #
        # * One item
        # ## Sub Item 1
        # ## Sub Item 2
        #

        # Ignore initial white spaces
        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Get the header and decide if it's a numbered or unnumbered list
        header = self._peek_token(MLTokenTypes.LIST)
        numbered = header.value[0] == "#"

        args, kwargs, tags = self._pop_arguments()

        # Parse all the following items
        nodes = self._process_list_nodes()

        self._save(
            ListNode(
                numbered, nodes, main_node=True, args=args, kwargs=kwargs, tags=tags
            )
        )

        return True

    def _process_list_nodes(self):
        # This parses all items of a list

        # Ignore initial white spaces
        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Parse the header and ignore the following white spaces
        header = self._get_token(MLTokenTypes.LIST).value
        self._get_token(BLTokenTypes.WHITESPACE)

        # Collect and parse the text of the item
        text = self._collect_text_content()
        content = self._parse_text_content(text)

        # Compute the level of the item
        level = len(header)

        nodes = []
        nodes.append(ListItemNode(str(level), content))

        while self._peek_token() not in [
            Token(BLTokenTypes.EOF),
            Token(BLTokenTypes.EOL),
        ]:
            # This is the SentenceNode inside the last node added to the list
            # which is used to append potential nested nodes
            last_node_sentence = nodes[-1].content

            # Ignore the initial white spaces
            with self:
                self._get_token(BLTokenTypes.WHITESPACE)

            if len(self._peek_token().value) == level:
                # The new item is on the same level

                # Get the header
                header = self._get_token().value

                # Ignore white spaces
                self._get_token(BLTokenTypes.WHITESPACE)

                # Collect and parse the text of the item
                text = self._collect_text_content()
                content = self._parse_text_content(text)

                nodes.append(ListItemNode(str(len(header)), content))
            elif len(self._peek_token().value) > level:
                # The new item is on a deeper level

                # Treat the new line as a new list
                numbered = self._peek_token().value[0] == "#"
                subnodes = self._process_list_nodes()

                last_node_sentence.content.append(ListNode(numbered, subnodes))
            else:
                break

        return nodes

    def _process_paragraph(self):
        # This parses a paragraph.
        # Paragraphs can be written on multiple lines and
        # end with an empty line.

        # Each line ends with EOL. This collects everything
        # until the EOL, then removes it. If the next token
        # is EOL we know that the paragraph is ended, otherwise
        # we continue to collect. If the token is EOF we
        # reached the end and we have to stop anyway.
        lines = []
        while self._peek_token() not in [
            Token(BLTokenTypes.EOL),
            Token(BLTokenTypes.EOF),
        ]:
            lines.append(
                self._collect_join(
                    [
                        Token(BLTokenTypes.EOL),
                    ]
                )
            )
            self._get_token(BLTokenTypes.EOL)

        text = " ".join(lines)
        sentence = self._parse_text_content(text)

        # Consume the arguments
        args, kwargs, tags = self._pop_arguments()

        self._save(ParagraphNode(sentence, args=args, kwargs=kwargs, tags=tags))

        return True



from mau.nodes.nodes import Node
from mau.nodes.page import PageNode


class FootnoteNode(Node):
    """A footnote created inside a piece of text."""

    node_type = "footnote"

    def __init__(
        self, content=None, number=None, reference_anchor=None, content_anchor=None
    ):
        super().__init__()
        self.content = content
        self.number = number
        self.reference_anchor = reference_anchor
        self.content_anchor = content_anchor

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "content": self.content,
            "number": self.number,
            "reference_anchor": self.reference_anchor,
            "content_anchor": self.content_anchor,
        }


class FootnotesEntryNode(FootnoteNode):
    """An entry of the list of footnotes."""

    node_type = "footnotes_entry"


class CommandFootnotesNode(PageNode):
    """This instructs Mau to insert the list of footnotes."""

    node_type = "command_footnotes"

    def __init__(self, entries, args=None, kwargs=None, tags=None):
        super().__init__(args, kwargs, tags)
        self.entries = entries

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "entries": self.entries,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }



from mau.lexers.text_lexer import TextLexer
from mau.nodes.footnotes import FootnoteNode
from mau.nodes.inline import (
    ClassNode,
    ImageNode,
    LinkNode,
    MacroNode,
    SentenceNode,
    StyleNode,
    TextNode,
    VerbatimNode,
)
from mau.nodes.references import ReferenceNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_empty_text():
    source = ""

    assert runner(source).nodes == []


def test_parse_word():
    source = "Word"

    expected = [
        SentenceNode(
            [
                TextNode("Word"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_multiple_words():
    source = "Many different words"

    expected = [
        SentenceNode(
            [
                TextNode("Many different words"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_parse_escape_word():
    source = r"\Escaped"

    expected = [
        SentenceNode(
            [
                TextNode("Escaped"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_parse_escape_symbol():
    source = r"\"Escaped"

    expected = [
        SentenceNode(
            [
                TextNode('"Escaped'),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_parse_style_underscore():
    source = "_Some text_"

    expected = [
        SentenceNode(
            [
                StyleNode(
                    "underscore",
                    SentenceNode(
                        [
                            TextNode("Some text"),
                        ]
                    ),
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_parse_style_star():
    source = "*Some text*"

    expected = [
        SentenceNode(
            [
                StyleNode(
                    "star",
                    SentenceNode(
                        [
                            TextNode("Some text"),
                        ]
                    ),
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_parse_style_caret():
    source = "^Some text^"

    expected = [
        SentenceNode(
            [
                StyleNode(
                    "caret",
                    SentenceNode(
                        [
                            TextNode("Some text"),
                        ]
                    ),
                ),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_parse_style_tilde():
    source = "~Some text~"

    expected = [
        SentenceNode(
            [
                StyleNode(
                    "tilde",
                    SentenceNode(
                        [
                            TextNode("Some text"),
                        ]
                    ),
                ),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_style_within_style():
    source = "_*Words with two styles*_"

    expected = [
        SentenceNode(
            [
                StyleNode(
                    "underscore",
                    SentenceNode(
                        [
                            StyleNode(
                                "star",
                                SentenceNode([TextNode("Words with two styles")]),
                            )
                        ]
                    ),
                )
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_paragraph_double_style_cancels_itself():
    source = "__Text__"

    expected = [
        SentenceNode(
            [
                StyleNode("underscore", SentenceNode([])),
                TextNode("Text"),
                StyleNode("underscore", SentenceNode([])),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_text_and_styles():
    source = "Some text _and style_ and *more style* here"

    expected = [
        SentenceNode(
            [
                TextNode("Some text "),
                StyleNode(
                    "underscore",
                    SentenceNode(
                        [
                            TextNode("and style"),
                        ]
                    ),
                ),
                TextNode(" and "),
                StyleNode(
                    "star",
                    SentenceNode(
                        [
                            TextNode("more style"),
                        ]
                    ),
                ),
                TextNode(" here"),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_parse_style_open():
    source = "_Text"

    expected = [
        SentenceNode(
            [
                TextNode("_Text"),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_verbatim():
    source = "`Many different words`"

    expected = [
        SentenceNode(
            [
                VerbatimNode("Many different words"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_verbatim_escaped_backtick():
    source = r"`\``"

    expected = [
        SentenceNode(
            [
                VerbatimNode("`"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_verbatim_style_inside_verbatim():
    source = r"`_Many different words_`"

    expected = [
        SentenceNode(
            [
                VerbatimNode("_Many different words_"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_verbatim_open():
    source = r"`Many different words"

    expected = [
        SentenceNode(
            [
                TextNode("`Many different words"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_verbatim_and_style():
    source = "Some text with `verbatim words` and _styled ones_"

    expected = [
        SentenceNode(
            [
                TextNode("Some text with "),
                VerbatimNode("verbatim words"),
                TextNode(" and "),
                StyleNode(
                    "underscore",
                    SentenceNode(
                        [
                            TextNode("styled ones"),
                        ]
                    ),
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_square_brackets():
    source = "This contains [ and ] and [this]"

    expected = [
        SentenceNode(
            [
                TextNode("This contains [ and ] and [this]"),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_collect_macro_arguments_single_argument():
    source = "(value1)"

    parser = init_parser(source)
    assert parser._collect_macro_args() == "value1"


def test_collect_macro_arguments_multiple_arguments():
    source = "(value1,value2)"

    parser = init_parser(source)
    assert parser._collect_macro_args() == "value1,value2"


def test_collect_macro_arguments_single_argument_with_quotes():
    source = '("value1")'

    parser = init_parser(source)
    assert parser._collect_macro_args() == '"value1"'


def test_collect_macro_arguments_single_argument_with_quotes_and_parenthesis():
    source = '("value1()")'

    parser = init_parser(source)
    assert parser._collect_macro_args() == '"value1()"'


def test_collect_macro_arguments_single_argument_with_parenthesis():
    source = "(value1())"

    parser = init_parser(source)
    assert parser._collect_macro_args() == "value1("


def test_collect_macro_arguments_multiple_argument_with_quotes_and_parenthesis():
    source = '("value1()",value2,value3)'

    parser = init_parser(source)
    assert parser._collect_macro_args() == '"value1()",value2,value3'


def test_collect_macro_arguments_multiple_argument_with_escaped_quotes():
    source = r"(\"value2,value3)"

    parser = init_parser(source)
    assert parser._collect_macro_args() == r"\"value2,value3"


def test_macro():
    source = "[macroname](value1,value2)"

    expected = [
        SentenceNode(
            [
                MacroNode(
                    "macroname",
                    args=["value1", "value2"],
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_incomplete_macro():
    source = "[macroname](value1"

    expected = [
        SentenceNode(
            [
                TextNode(
                    "[macroname](value1",
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_arguments_with_quotes():
    source = '[macroname]("value1,value2")'

    expected = [
        SentenceNode(
            [
                MacroNode(
                    "macroname",
                    args=["value1,value2"],
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_named_arguments():
    source = "[macroname](name,arg1=value1)"

    expected = [
        SentenceNode(
            [
                MacroNode(
                    "macroname",
                    args=["name"],
                    kwargs={"arg1": "value1"},
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_link():
    source = '[link](https://somedomain.org/the/path, "link text")'

    expected = [
        SentenceNode(
            [
                LinkNode("https://somedomain.org/the/path", "link text"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_link_without_text():
    source = '[link]("https://somedomain.org/the/path")'

    expected = [
        SentenceNode(
            [
                LinkNode(
                    "https://somedomain.org/the/path", "https://somedomain.org/the/path"
                ),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_mailto():
    source = "[mailto](info@projectmau.org)"

    expected = [
        SentenceNode(
            [
                LinkNode("mailto:info@projectmau.org", "info@projectmau.org"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_mailto_custom_text():
    source = '[mailto](info@projectmau.org, "my email")'

    expected = [
        SentenceNode(
            [
                LinkNode("mailto:info@projectmau.org", "my email"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_footnote():
    source = "[footnote](notename)"

    footnote_node = FootnoteNode()
    expected = [SentenceNode([footnote_node])]

    parser = runner(source)
    assert parser.nodes == expected
    assert parser.footnotes == {"notename": footnote_node}


def test_macro_reference():
    source = "[reference](solution, value)"

    reference_node = ReferenceNode("solution", "value")
    expected = [SentenceNode([reference_node])]

    parser = runner(source)
    assert parser.nodes == expected
    assert parser.references == {
        ("solution", "value"): reference_node,
    }


def test_single_class():
    source = 'Some text [class]("text with that class", classname)'

    expected = [
        SentenceNode(
            [
                TextNode("Some text "),
                ClassNode(
                    ["classname"],
                    SentenceNode(
                        [
                            TextNode("text with that class"),
                        ]
                    ),
                ),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_multiple_classes():
    source = 'Some text [class]("text with that class", "classname1,classname2")'

    expected = [
        SentenceNode(
            [
                TextNode("Some text "),
                ClassNode(
                    ["classname1", "classname2"],
                    SentenceNode(
                        [
                            TextNode("text with that class"),
                        ]
                    ),
                ),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_parse_class_with_rich_text():
    source = '[class]("Some text with `verbatim words` and _styled ones_", classname)'

    expected = [
        SentenceNode(
            [
                ClassNode(
                    ["classname"],
                    SentenceNode(
                        [
                            TextNode("Some text with "),
                            VerbatimNode("verbatim words"),
                            TextNode(" and "),
                            StyleNode(
                                "underscore", SentenceNode([TextNode("styled ones")])
                            ),
                        ]
                    ),
                ),
            ]
        )
    ]

    assert runner(source).nodes == expected


def test_macro_image():
    source = "[image](/the/path.jpg)"

    expected = [
        SentenceNode(
            [
                ImageNode("/the/path.jpg"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_image_with_alt_text():
    source = '[image](/the/path.jpg,"alt name")'

    expected = [
        SentenceNode(
            [
                ImageNode("/the/path.jpg", alt_text="alt name"),
            ]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_image_with_width_and_height():
    source = "[image](/the/path.jpg,width=1200,height=600)"

    expected = [
        SentenceNode(
            [
                ImageNode("/the/path.jpg", alt_text=None, width="1200", height="600"),
            ]
        ),
    ]

    assert runner(source).nodes == expected



from mau.nodes.page import PageNode


class TocNode(PageNode):
    """A Table of Contents.

    This node contains the entries of the Table of Contents.
    """

    node_type = "toc"

    def __init__(self, entries, args=None, kwargs=None, tags=None):
        super().__init__(args, kwargs, tags)
        self.entries = entries

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "entries": self.entries,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class TocEntryNode(PageNode):
    """An entry of the Table of Contents.

    This node contains an entry of the Table of Contents.
    """

    node_type = "toc_entry"

    def __init__(self, value, anchor, children=None, args=None, kwargs=None, tags=None):
        super().__init__(args, kwargs, tags)
        self.value = value
        self.anchor = anchor
        self.children = children or []

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "value": self.value,
            "anchor": self.anchor,
            "children": self.children,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }



from mau.nodes.arguments import NamedArgumentNode, UnnamedArgumentNode


def test_unnamed_argument_node_value():
    node = UnnamedArgumentNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "unnamed_argument"
    assert node == UnnamedArgumentNode("somevalue")


def test_named_argument_node_value():
    node = NamedArgumentNode("somekey", "somevalue")

    assert node.key == "somekey"
    assert node.value == "somevalue"
    assert node.node_type == "named_argument"
    assert node == NamedArgumentNode("somekey", "somevalue")
